#include "AAOIntegralPrecon.hpp"
class DGVAcousticRHSAssemblerWavefieldForAAOPrecon : public DGVAcousticRHSAssembler {
    const Vectors &U;
    mutable int step = 1;
    mutable bool increase = false;
    double dt;
    bool fwd = true;
public:
    explicit DGVAcousticRHSAssemblerWavefieldForAAOPrecon(const Vectors &U_, double dt_, bool fwd_)
            : DGVAcousticRHSAssembler(),
              U(U_), dt(dt_), fwd(fwd_) {
        step = fwd ? 1 : U.size() - 2;
    }

    void RHS(double t, Vector &b) const override {
        int stepTwo = fwd ? step - 1 : step + 1;
        b = (1.0 / dt) * (U[step] - U[stepTwo]);
        if (increase) {
            increase = false;
            if (fwd) step++;
            else step--;
        } else {
            increase = true;
        }
    }

    void Initialize(Vector &) const override {}

    ~DGVAcousticRHSAssemblerWavefieldForAAOPrecon() = default;

};

void AAOIntegralPrecon::multiply_transpose(Vectors &u, const Vectors &b) const {
    multiply2(u, b, false);
}

void AAOIntegralPrecon::multiply(Vectors &u, const Vectors &b) const {
    multiply2(u, b, true);
}

void AAOIntegralPrecon::multiply2(Vectors &u, const Vectors &b, bool fwd) const {
    const std::shared_ptr<DGVAcousticAssembleCombined> assemble = fwd ? assembleFWD : assembleBWD;
    auto RHSWave = std::make_shared<DGVAcousticRHSAssemblerWavefieldForAAOPrecon>(b, assemble->StepSize(), fwd);
    auto abch = std::shared_ptr<ABCHandler>(nullptr);
    FWITimeSolver TSolveR(WaveVec, abch, CRANK_NICOLSON_HARDCODED);
    assemble->SetRhsAssembler(RHSWave);
    assemble->SetEleCache(&eleCache);
    Vector tmp2 = (*solvIM) * b[fwd ? 0 : b.size() - 1];
    tmp2.MakeAdditive();
    assemble->DefineInitialValue(&tmp2);
    TSolveR.Solve(assemble, u);
    if (!fwd) u.reverse();
    assembleFWD->ResetTime(t0, T, dt);
    assembleBWD->ResetTime(t0, T, dt);
}

AAOIntegralPrecon::AAOIntegralPrecon(const Vector &WaveVec_,
                                     ABCHandlerAAO *abch_,
                                     const EleCacheType &eleCache, const Vector &MatVec,
                                     const FWIVAcousticInversionProb &prob, double t0, double T,
                                     double dt)
    : WaveVec(WaveVec_), Mass(WaveVec), eleCache(eleCache), t0(t0), dt(dt), T(T) {
  auto Mat = std::make_unique<VAcousticMatFromVec>(prob, MatVec, nullptr,abch_);
    auto matrixAssemblerFWD = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), true, &eleCache);
    auto Mat2 = std::make_unique<VAcousticMatFromVec>(prob, MatVec,nullptr, abch_);
    auto matrixAssemblerBWD = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat2), false, &eleCache);
    auto VAPlotter = std::make_shared<DGVAcousticPlotter>(prob.GetNumL(), &eleCache);
    assembleFWD = std::make_shared<DGVAcousticAssembleCombined>(VAPlotter, matrixAssemblerFWD,
                                                                nullptr, t0, T, dt);
    assembleBWD = std::make_shared<DGVAcousticAssembleCombined>(VAPlotter, matrixAssemblerBWD,
                                                                nullptr, t0, T, dt);
    matrixAssemblerFWD->MassMatrix(Mass);
    solvIM = std::make_unique<LinearSolver>(GetPC("PointBlockJacobi"));
    (*solvIM)(Mass, 1);

}
