//
// Created by cr on 13.10.21.
//

#include "FWILinearizedAssembler.hpp"
void FWILinearizedAssembler::assembleLinearizedMGalerkin(const Vector &materialDirection,
                                                         Matrix &mPrime) const {
  mPrime = 0.0;
  for (const auto &elemIt: eleCache) {
    const DGAcousticElement *elem = elemIt.second;
    const Cell &C = elem->GetCell();
    DGRowEntries mPrime_c(mPrime, C, C);
    int Row = materialPoint.find_row(C()).Id();
    double rhoFac= getRhoFac(Row);
    double rhoDirec = materialDirection(Row, 0);
    std::vector<double> pFac = getpFacs(materialDirection, Row, C);
    int nV = elem->V_dimension();
    int nP = elem->shape_size();
    int p_dim = elem->p_dimension();
    for (int q = 0; q < elem->nQ(); ++q) {
      double w = elem->QWeight(q);
      for (int i = 0; i < nV; ++i) {
        VectorField V_i = elem->Velocity(q, i);
        for (int j = 0; j < nV; ++j) {
          VectorField V_j = elem->Velocity(q, j);
          mPrime_c(i, j) += w * V_i * V_j * rhoDirec*rhoFac;

        }
      }
    }
    for (int i = 0; i < nP; ++i) {
      for (int q = 0; q < elem->nQ(); ++q) {
        double w = elem->QWeight(q);
        double P_i = elem->Pressure(q, i);
        for (int j = 0; j < nP; ++j) {
          double P_j = elem->Pressure(q, j);
          mPrime_c(nV + i, nV + j) += w * P_i * P_j * pFac[0];
          for (int n = 1; n < p_dim; ++n) {
            mPrime_c(nV + n * nP + i, nV + n * nP + j) += w * P_i * P_j * pFac[n];
          }
        }
      }
    }
  }
//  mPrime.CreateSparse();
}

void FWILinearizedAssembler::assembleDGalerkin2(const Vector &materialDirection,
                                                Matrix &diag) const {
  diag = 0.0;
  for (const auto &elemIt: eleCache) {
    auto elem = elemIt.second;
    const Cell &C = elem->GetCell();
    DGRowEntries diag_c(diag, C, C);
    int Row = materialPoint.find_row(elemIt.first).Id();
    std::vector<double> pFac = getpFacs(materialDirection, Row, C);
    int nV = elem->V_dimension();
    int nP = elem->shape_size();
    int p_dim = elem->p_dimension();
    for (int q = 0; q < elem->nQ(); ++q) {
      double w = elem->QWeight(q);
      for (int i = 0; i < nP; ++i) {
        double P_i = elem->Pressure(q, i);
        for (int j = 0; j < nP; ++j) {
          double P_j = elem->Pressure(q, j);
          for (int n = 1; n < p_dim; ++n) {
            diag_c(nV + n * nP + i, nV + n * nP + j) +=
                w * P_i * P_j * pFac[n] * 1.0 / getTauI(C, n);
          }
        }
      }
    }
  }
}

void FWILinearizedAssembler::MPrimeMatVecOneElemDiagonal(const DGAcousticElement &elem,
                                                         const Vector &materialDirection,
                                                         const Vector &JVec,
                                                         const Vector &tmp,
                                                         Vector &Vec) const {
  const Cell &C = elem.GetCell();
  int Row = materialPoint.find_row(C()).Id();
  double rhoFac = getRhoFac(Row);
  double rhoDirec = materialDirection(Row, 0);

  std::vector<double> pFac = getpFacs(materialDirection, Row, C);
  int nV = elem.V_dimension();
  int nP = elem.shape_size();
  int p_dim = elem.p_dimension();
  int Row2 = elem.Row().Id();
//  int Row2 = Vec.find_row(C()).Id();
  for (int j = 0; j < nV; ++j) {
    Vec(Row2, j) = rhoDirec * rhoFac * tmp(Row2, j);
  }
  for (int j = 0; j < nP; ++j) {
    Vec(Row2, j + nV) = tmp(Row2, j + nV) * pFac[0];
    for (int n = 1; n < p_dim; ++n) {
      Vec(Row2, j + nV + n * nP) = pFac[n] * (tmp(Row2, j + nV + n * nP) +
                                              1.0 / problem_fwd->Tau_i(C, C(), n) *
                                              JVec(Row2, j + nV + n * nP));
    }
  }

}

void FWILinearizedAssembler::MPrimeMatVecOneElem(const DGAcousticElement &elem,
                                                 const Vector &materialDirection,
                                                 const Vector &JVec,
                                                 const Vector &tmp,
                                                 Vector &Vec) const {
  const Cell &C = elem.GetCell();
  int Row = materialPoint.find_row(C()).Id();
  double rhoFac = getRhoFac(Row);
  double rhoDirec = materialDirection(Row, 0);
  std::vector<double> pFac = getpFacs(materialDirection, Row, C);
  int nV = elem.V_dimension();
  int nP = elem.shape_size();
  int p_dim = elem.p_dimension();
  int Row2 = elem.Row().Id();
  for (int q = 0; q < elem.nQ(); ++q) {
    double w = elem.QWeight(q);
    VectorField VVec = elem.Velocity(q, tmp);
    for (int j = 0; j < nV; ++j) {
      VectorField V_j = elem.Velocity(q, j);
      Vec(Row2, j) += w * VVec * V_j * rhoDirec * rhoFac;
    }
  }
  for (int q = 0; q < elem.nQ(); ++q) {
    double w = elem.QWeight(q);
    for (int j = 0; j < nP; ++j) {
      double PVec = elem.Pressure_i(q, tmp, 0);
      double P_j = elem.Pressure(q, j);
      Vec(Row2, j + nV) += w * PVec * P_j * pFac[0];
      for (int n = 1; n < p_dim; ++n) {
        double PVecn = elem.Pressure_i(q, tmp, n);
        double PJVec = elem.Pressure_i(q, JVec, n);
        Vec(Row2, j + nV + n * nP) += w * P_j * pFac[n]
                                      * (PVecn + 1.0 / problem_fwd->Tau_i(C, C(), n) * PJVec);
      }
    }
  }
}

void FWILinearizedAssembler::MPrimeMatVec(const Vector &materialDirection,
                                          const Vector &JVec, Vector &Vec) const {
  Vector tmp(Vec);
  Vec.Clear();
  for (const auto &elemIt: eleCache) {
    MPrimeMatVecOneElem(*elemIt.second, materialDirection, JVec, tmp, Vec);
  }
}

void FWILinearizedAssembler::MPrimeMatVecDiagonal(const Vector &materialDirection,
                                                  const Vector &JVec, Vector &Vec) const {
  Vector tmp(Vec);
  Vec.Clear();
  for (const auto &elemIt: eleCache) {
    MPrimeMatVecOneElemDiagonal(*elemIt.second, materialDirection, JVec, tmp, Vec);
  }
}


std::vector<double> FWILinearizedAssembler::getpFacs(const Vector &materialDirection,
                                                     int Row, const Cell &C) const {

    if (problem_fwd->getParametrization() == "taup_positive") {
        double taup_direc = (numL > 0) ? materialDirection(Row, 2) : 0;
        return pFacsTaupPositive(getRho(Row), getVP(Row), getTauP(Row),
                                 materialDirection(Row, 0), materialDirection(Row, 1),
                                 taup_direc,
                                 getAlpha(C), numL);
    } else if (problem_fwd->getParametrization() == "taup") {
        double taup_direc = (numL > 0) ? materialDirection(Row, 2) : 0;
        return pFacsTaup(getRho(Row), getVP(Row), getTauP(Row),
                         materialDirection(Row, 0),
                         materialDirection(Row, 1),
                         taup_direc,
                         getAlpha(C), numL);
    } else if (problem_fwd->getParametrization() == "native") {
        std::vector<double> vec;
        for (int i = 0; i < numL + 1; ++i) {
            double kappa_i = materialPoint(Row, 1 + i);
            double kappa_iDirec = materialDirection(Row, 1 + i);
            vec.push_back(-(1.0/(kappa_i * kappa_i)) * kappa_iDirec);
        }
        return vec;
        Exit("not implemented!");
    } else {
        Exit("wrong parametrization");
    }
}

double FWILinearizedAssembler::getAlpha(const Cell &C) const {
  return (numL == 0) ? 0.0 : problem_fwd->alpha(C, C());
}

double FWILinearizedAssembler::getTauI(const Cell &C, int n) const {
  return problem_fwd->Tau_i(C, C(), n);
}
double FWILinearizedAssembler::getRhoFac(int Row) const {
  if (problem_fwd->getParametrization() == "taup_positive") {
    return exp(materialPoint(Row, 0));
  } else if (problem_fwd->getParametrization() == "taup" ||
             problem_fwd->getParametrization() == "native") {
    return 1.0;
  } else {
    Exit("wrong parametrization");
  }
}

double FWILinearizedAssembler::getRho(int Row) const {
  if (problem_fwd->getParametrization() == "taup_positive") {
    return exp(materialPoint(Row, 0));
  } else if (problem_fwd->getParametrization() == "taup") {
    return materialPoint(Row, 0);
  } else if (problem_fwd->getParametrization() == "native") {
    return materialPoint(Row, 0);
  } else {
    Exit("wrong parametrization");
  }

}

double FWILinearizedAssembler::getVP(int Row) const {
  if (problem_fwd->getParametrization() == "taup_positive") {
    return exp(materialPoint(Row, 1));
  } else if (problem_fwd->getParametrization() == "taup") {
    return materialPoint(Row, 1);
  } else {
    Exit("wrong parametrization");
  }
}

double FWILinearizedAssembler::getTauP(int Row) const {
  if (problem_fwd->getParametrization() == "taup_positive") {
    return (numL == 0) ? 0.0 : exp(materialPoint(Row, 2));
  } else if (problem_fwd->getParametrization() == "taup") {
    return (numL == 0) ? 0.0 : materialPoint(Row, 2);
  } else {
    Exit("wrong parametrization");
  }
}

std::vector<double>
pFacsTaupPositive(double rho, double vP, double tauP, double rhoDirec, double vPDirec,
                  double tauPDirec,
                  double alpha, int numL) {
  double tauPFac = (numL == 0) ? 0.0 : alpha * tauP / (Squared(vP) * rho) * tauPDirec;
  double rhoFac = (numL == 0) ? (-1.0 / (Squared(vP) * rho) * rhoDirec)
                              : -(1.0 + alpha * tauP) / (Squared(vP) * rho) * rhoDirec;

  double vPFac = (numL == 0) ? -2.0 / (Squared(vP) * rho) * vPDirec
                             : -2.0 * (1.0 + alpha * tauP) / (Squared(vP) * rho) * vPDirec;
  double tau_pfac2 = (numL == 0) ? 0.0 : -tauPDirec / (Squared(vP) * rho * tauP);
  std::vector<double> vec;
  vec.push_back(rhoFac + vPFac + tauPFac);
  if (numL == 0) {
    vec.push_back(0.0);
  } else {
    for (int i = 0; i < numL; ++i) {
      vec.push_back(rhoFac / tauP + vPFac / tauP + tau_pfac2);
    }
  }
  return vec;
}

std::vector<double>
pFacsTaup(double rho, double vP, double tauP, double rhoDirec, double vPDirec, double tauPDirec,
          double alpha, int numL) {
  double tauPFac = (numL == 0) ? 0.0 : alpha / (Squared(vP) * rho) * tauPDirec;
  double rhoFac = (numL == 0) ? (-1.0 / (Squared(vP * rho)) * rhoDirec)
                              : -(1 + alpha * tauP) / (Squared(rho * vP)) * rhoDirec;

  double vPFac = (numL == 0) ? -2.0 / (vP * vP * vP * rho) * vPDirec
                             : -2.0 * (1.0 + alpha * tauP) / (vP * vP * vP * rho) * vPDirec;
  double tau_pfac2 = (numL == 0) ? 0.0 : -tauPDirec / (Squared(vP * tauP) * rho);
  std::vector<double> vec;
  vec.push_back(rhoFac + vPFac + tauPFac);
  if (numL == 0) {
    vec.push_back(0.0);
  } else {
    for (int i = 0; i < numL; ++i) {
      vec.push_back(rhoFac / tauP + vPFac / tauP + tau_pfac2);
    }
  }
  return vec;
}