//
// Created by cr on 13.10.21.
//

#ifndef FWI_FWILINEARIZEDASSEMBLER_HPP
#define FWI_FWILINEARIZEDASSEMBLER_HPP

#include "DGAcousticElement.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "Helper/FWIMppHelpers.hpp"
#include "Problems/FWIAcousticInversionProb.hpp"
#include "lib4_fem/algebra/Vector.hpp"

std::vector<double>
pFacsTaupPositive(double rho, double vP, double tauP, double rhoDirec, double vPDirec, double tauPDirec,
                  double alpha, int numL);

std::vector<double> pFacsTaup(double rho, double vP, double tauP, double rhoDirec, double vPDirec, double tauPDirec,
                              double alpha, int numL);

class FWILinearizedAssembler {

  void MPrimeMatVecOneElem(const DGAcousticElement &elemm,
                             const Vector &materialDirection,
                             const Vector &JVec,
                             const Vector &tmp,
                             Vector &Vec) const;

    double getTauP(int Row) const;

    double getVP(int Row) const;

    double getRho(int Row) const;

  double getRhoFac(int Row) const;

  double getTauI(const Cell &C, int n) const;

  double getAlpha(const Cell &C) const;

  std::vector<double> getpFacs(const Vector &materialDirection,
                               int Row, const Cell &C) const;


  std::shared_ptr<FWIVAcousticInversionProb> problem_fwd;
    const std::unordered_map<Point, DGAcousticElement *> &eleCache;
    int numL = -1;
  const Vector &materialPoint;

public:
    FWILinearizedAssembler(std::shared_ptr<FWIVAcousticInversionProb> problem_fwd_,
                           const std::unordered_map<Point, DGAcousticElement *> &eleCache_,
                           const Vector &materialPoint_) :
            eleCache(eleCache_),
            problem_fwd(problem_fwd_),
            numL(problem_fwd_->GetNumL()),
            materialPoint(materialPoint_) {}

  void assembleLinearizedMGalerkin(const Vector &materialDirection, Matrix &mPrime) const;

    void MPrimeMatVec(const Vector &materialDirection, const Vector &JVec, Vector &Vec) const;

    void MPrimeMatVecDiagonal(const Vector &materialDirection, const Vector &JVec, Vector &Vec) const;

  void MPrimeMatVecOneElemDiagonal(const DGAcousticElement &elem,
                                     const Vector &materialDirection,
                                     const Vector &JVec,
                                     const Vector &tmp,
                                     Vector &Vec) const;

    void assembleDGalerkin2(const Vector &materialDirection, Matrix &diag) const;

    double GetZK(const Cell& C) const {
      if (problem_fwd->getParametrization() == "native") {
        double rhoK = materialPoint(C(), 0);
        double kappa = 0;
        for (int i = 0; i < numL+1; ++i) {
           kappa += materialPoint(C(), 1+i);
        }
        return sqrt(kappa*rhoK);
      }else if (problem_fwd->getParametrization() == "taup") {
        double alpha = (numL == 0) ? 1 : problem_fwd->alpha(C, C());
        double rhoK = materialPoint(C(), 0);
        double vK = materialPoint(C(), 1);
        double taupK = (numL == 0) ? 0 : materialPoint(C(), 2);
        return zKTau(numL, alpha, rhoK, vK, taupK);
      } else if (problem_fwd->getParametrization() == "taup_positive") {
        double alpha = (numL == 0) ? 1 : problem_fwd->alpha(C, C());
        double rhoK = materialPoint(C(), 0);
        double vK = materialPoint(C(), 1);
        double taupK = (numL == 0) ? 0 : materialPoint(C(), 2);
        return zKTauPositive(numL, alpha, rhoK, vK, taupK);
      } else {
        Exit("Wrong Parametrization");
      }
    }
  double GetZKPrime(const Cell& C, const Vector &materialDirection) const {
    if (problem_fwd->getParametrization() == "native") {
      double rhoK = materialPoint(C(), 0);
      double kappa = 0;
      for (int i = 0; i < numL+1; ++i) {
        kappa += materialPoint(C(), 1+i);
      }
      double ZK = sqrt(kappa*rhoK);
      double ret = 0.0;
      for (int i = 0; i < numL+1; ++i) {
        ret += rhoK * materialDirection(C(), 1+i);
      }
      ret += materialDirection(C(), 0) * kappa;
      return 0.5*(ret/sqrt(kappa*rhoK));
    } else if (problem_fwd->getParametrization() == "taup"){
        double alpha = (numL == 0) ? 1 : problem_fwd->alpha(C, C());
        double rhoK = materialPoint(C(), 0);
        double vK = materialPoint(C(), 1);
        double taupK = (numL == 0) ? 0 : materialPoint(C(), 2);
        double rhoDK = materialDirection(C(), 0);
        double vDK = materialDirection(C(), 1);
        double taupDK = (numL == 0) ? 0 : materialDirection(C(), 2);
        return zKPrimeTau(numL, alpha, rhoK, vK, taupK, rhoDK, vDK, taupDK);
      }else if (problem_fwd->getParametrization() == "taup_positive") {
      double alpha = (numL == 0) ? 1 : problem_fwd->alpha(C, C());
      double rhoK = materialPoint(C(), 0);
      double vK = materialPoint(C(), 1);
      double taupK = (numL == 0) ? 0 : materialPoint(C(), 2);
      double rhoDK = materialDirection(C(), 0);
      double vDK = materialDirection(C(), 1);
      double taupDK = (numL == 0) ? 0 : materialDirection(C(), 2);
      return zKPrimeTauPositive(numL, alpha, rhoK, vK, taupK, rhoDK, vDK, taupDK);
    } else {
      Exit("Wrong Parametrization");
    }
  }

};


#endif //FWI_FWILINEARIZEDASSEMBLER_HPP
