//#include "m++.hpp"
#include "FWINonlinearSolvers.hpp"
#include "FWIProblemWrapper.hpp"
#include "ForwardModels/ForwardModelRed.hpp"
#include "Helper/VTUModelReader.hpp"
#include "Helper/VectorFillFunctions.hpp"
#include "MeshIndex.hpp"
#include "Problems/FWIInversionProblems.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "Observation/ObservationCreator.hpp"
#include "m++.hpp"

int FWIConv() {
  std::vector<Vectors> U_ms;
  auto meshes = getFWIMesh();
  WPConf Conf;
  int degStart = 0;
  int degEnd = 1;
  Config::Get("degStart", degStart);
  Config::Get("degEnd", degEnd);
  Config::Get("Problem",Conf.s_prob );
  std::vector<DGSeismogram> s_ms;
  auto problem = GetInversionProblem(Conf);
  int nrLvs = meshes->Level() - meshes->pLevel();
  std::string l;
  std::shared_ptr<ObservationSpecification> spec;
  double signalFreq = Conf.SConf.SignalFreq;
  for (int deg = degStart; deg <= degEnd; ++deg) {
    for (int lv = meshes->pLevel(); lv <= meshes->Level(); ++lv) {
      MeshIndex lev{lv, -1};
      FWIDiscretizationStack Stack(meshes, problem->GetParamComps(),
                                   problem->GetStateComps(), deg,
                                   lev);
      (*Stack.wD)().PrintInfo();
      (*meshes)[lev].PrintInfo();
      auto eleCache = std::make_unique<EleCacheContainer>(*Stack.wV, problem->GetNumL());
      Config::Get("Specfile",l);
      if (l.empty()) {
        spec = ObservationSpecificationBuilder().WithT(Conf.TConf.T).Withdt(Conf.TConf.dt).Witht0(Conf.TConf.t0).CreateShared();
      }else{
        Exit("USE JSON FORMAT");
        int SamplingFactor = 1;
        Config::Get("SpecfileSamplingFactor",SamplingFactor);
      }
      FWIObservationOperatorDGTS *Obs =
          ObservationCreator(eleCache->Get(),
                                                              Stack.GetMesh(),
                                problem->GetNumL()).WithSpec(
          spec).Create();
      std::cout << std::setprecision(15);
      problem->setInitialValueFlag(true);
      auto M = std::make_unique<Vector>(*Stack.mV);
      auto M2 = std::make_unique<Vector>(*Stack.mV);
      M2->Clear();
      M->Clear();
      FillVectorFromProblem(*problem, *M2);
      ReparametrizeMaterialVector(*M2, *problem, *M);
      auto PMLH = std::make_shared<ABCHandler>();
      problem->SetShot(0);
      FWITimeSolver TSolveR(*Stack.wV, PMLH, Conf.TConf.rkorder);
      auto Mat = std::make_unique<VAcousticMatFromVec>(*problem, *M2);
      auto Mat2 = std::make_unique<VAcousticMatFromVec>(*problem, *M2);
      auto MatAss = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), true,
                                                                 eleCache->Get().get());
      auto plotter = std::make_shared<DGVAcousticPlotter>(problem->GetNumL(), eleCache->Get().get());
      auto problemVA = std::make_shared<FWIProblemWrapper>(*problem, *problem, std::move(M),
                                                problem->GetNumL(),
                                                problem->GetDim(),
                                                             problem);

      auto rhsAssembler = std::make_shared<DGVAcousticRHSAssemblerDefault>(*problemVA);
      auto assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(plotter, MatAss, rhsAssembler,
                                                        Conf.TConf.t0, Conf.TConf.T, Conf.TConf.dt);
      Vectors U_m(Conf.TConf.timeSteps + 1, *Stack.wV);
      DGSeismogram s_m(Obs->GetSpaceOp().GetParaDat());
      assembleCombined->SetEleCache(eleCache->Get().get());
      assembleCombined->SetInitialValue(U_m[0]);
      TSolveR.SolveAndMeasure(assembleCombined, *Obs, s_m, &U_m);
      s_ms.push_back(s_m);
      //      s_m.collect();
      //      s_m.writeToFile(l+"S__deg_" +  std::to_string(deg)+"__lv_" +
      //      std::to_string(lv) + "__freq_" +  std::to_string(signalFreq));
    }
  }
  for (int i = 1; i < s_ms.size();++i){
    DGSeismogram diff = s_ms[i-1];
    diff -= s_ms[i];
    MOUT(sqrt(diff * diff));
  }
  return 1;
}
