//
// Created by cr on 02.12.21.
//

#include "FWITaper.hpp"

#include <cmath>
#include <utility>

#include "Cell.hpp"
#include "Helper/FWIHelpers.hpp"
#include "Problems/FWIAcousticInversionProb.hpp"
#include "Point.hpp"
#include "Vector.hpp"
#include "lib6_app/seismogram/ObservationSpecification.hpp"

FWITaper::FWITaper(const Vector &Vec, const ObservationSpecification &spec, TaperConf Conf_, FWIVAcousticInversionProb &prob)
    : taperVector(0.0, Vec), Conf(std::move(Conf_)) {
  taperVector = 1.0;
  paramComps = prob.GetParamComps();
  for (cell c = taperVector.cells(); c != taperVector.cells_end(); ++c) {
    if (c()[0] < Conf.boxLeft || c()[0] > Conf.boxRight || c()[1] < Conf.boxLower || c()[1] > Conf.boxUpper) {
      taperVector(c(), 0) = 0.0;
      continue;
    }
    bool found = false;
    if (Conf.taperIgnoreReceiverCells) {
      for (const auto &re: spec) {
        if (c.PointInCell(re)) {
          found = true;
          break;
        }
      }
      if (found) {
        taperVector(c(), 0) = 0.0;
        continue;
      }
    }
    if (Conf.taperIgnoreSourceCells) {
      found = false;
      for (int shot = 0; shot < prob.GetShotCount(); ++shot) {
        prob.SetShot(shot);
        int nrPointSrc = prob.getNrPointSrc();
        for (int i = 0; i < nrPointSrc; ++i) {
          Point src = prob.getPointSrcLoc(i);
          if (c.PointInCell(src)) {
            found = true;
            break;
          }
        }
      }
      if (found) {
        taperVector(c(), 0) = 0.0;
        continue;
      }
    }
    double tapeval = 0.0;
    int insupp = 0;
    for (int shot = 0; shot < prob.GetShotCount(); ++shot) {
      prob.SetShot(shot);
      int nrPointSrc = prob.getNrPointSrc();
      for (int i = 0; i < nrPointSrc; ++i) {
        Point src = prob.getPointSrcLoc(i);
        double distance = dist(c(), src) / Conf.taperRadiusSource;
        if (distance > 1.0) { continue;
        } else {
          tapeval += taperFunc(distance);
          insupp++;
        }
      }
    }
    for (const auto &re: spec) {
      double distance = dist(c(), re) / Conf.taperRadiusReceiver;
      if (distance > 1.0) { continue;
      }
      tapeval += taperFunc(distance);
      insupp++;
    }
    if (insupp > 0) {
      taperVector(c(), 0) = 1 - tapeval / insupp;
    }
  }
}

void FWITaper::ApplyTaperVector(Vector &Mat) {
  for (cell c = Mat.cells(); c != Mat.cells_end(); ++c) {
    for (int i = 0; i < paramComps; ++i) {
      Mat(c(), i) *= taperVector(c(), 0);
    }
  }
}

double FWITaper::taperFunc(double distance) const {
  return exp(Conf.taperExponent * (1 / (1 - pow(distance, 4)) - 1));
}
