//
// Created by cr on 02.12.21.
//

#ifndef FWI_FWITAPER_HPP
#define FWI_FWITAPER_HPP

#include <utility>
#include <vector>

#include "Cell.hpp"
#include "Config.hpp"
#include "ParameterImage.hpp"
#include "Problems/FWIAcousticInversionProb.hpp"
#include "Vector.hpp"
#include "lib6_app/seismogram/ObservationSpecification.hpp"

struct TaperConf {
  double taperRadiusSource = -1.0;
  double taperRadiusReceiver = -1.0;
  double boxUpper = infty;
  double boxLower = -infty;
  double boxLeft = -infty;
  double boxRight = infty;
  double taperExponent = -5;
  bool taperIgnoreReceiverCells = true;
  bool taperIgnoreSourceCells = true;
  std::vector<double> MinMaxValues{-1.0, -1.0, -1.0, -1.0, -1.0, -1.0};
  bool UseMinMaxTaper = false;
  TaperConf() {
    Config::Get("taperRadiusSource", taperRadiusSource);
    Config::Get("taperBoxUpper", boxUpper);
    Config::Get("taperBoxLower", boxLower);
    Config::Get("taperBoxLeft", boxLeft);
    Config::Get("taperBoxRight", boxRight);
    Config::Get("taperRadiusReceiver", taperRadiusReceiver);
    Config::Get("taperExponent", taperExponent);
    Config::Get("taperIgnoreSourceCells", taperIgnoreSourceCells);
    Config::Get("taperIgnoreReceiverCells", taperIgnoreReceiverCells);
    Config::Get("taperRhoMin", MinMaxValues[0]);
    Config::Get("taperRhoMax", MinMaxValues[1]);
    Config::Get("taperVpMin", MinMaxValues[2]);
    Config::Get("taperVpMax", MinMaxValues[3]);
    Config::Get("taperTaupMin", MinMaxValues[4]);
    Config::Get("taperTaupMax", MinMaxValues[5]);
    for (double v : MinMaxValues) {
      if (v > 0) {
        UseMinMaxTaper = true;
      }
    }
  }
};

class FWITaper {
  Vector taperVector;
  int paramComps = -1;

  [[nodiscard]] double taperFunc(double distance) const;

public:
  TaperConf Conf;

  explicit FWITaper(const Vector &Vec, const ObservationSpecification &spec,
                    TaperConf Conf_, FWIVAcousticInversionProb &prob);

  explicit FWITaper(Vector Vec,const std::string &fn,const std::string &readtype, int paramComps_)
      : taperVector(std::move(Vec)), paramComps(paramComps_) {
    if (readtype == "png") {
      ParameterImage Parameters("Taper", fn, std::pair<double, double>(Conf.boxLeft, Conf.boxRight),
                         std::pair<double, double>(Conf.boxLower, Conf.boxUpper));
      for (cell c = taperVector.cells(); c != taperVector.cells_end(); ++c) {
        taperVector(c(), 0) = Parameters.ValueAtPoint(c());
      }
    }else if (readtype == "JSON"){
      taperVector.Clear();
      ReadJSONToVec(fn,taperVector,0);

    }else{
      THROW("Read type unkown!")
    }
  }

  const Vector &GetTaperVector() { return taperVector; }
  void ApplyTaperVector(Vector &Mat);
  void ApplyMinMaxTaperAddition(Vector &Iterate, Vector &Update,
                                const FWIVAcousticInversionProb &prob) const {
    if (Conf.UseMinMaxTaper) {
      auto ValsTmp = Conf.MinMaxValues;
      if (prob.getParametrization() == "taup_positive") {
        for (double &i : ValsTmp) {
          i == -1 ? -1 : i = log(i);
        }
      }
      for (cell c = Iterate.cells(); c != Iterate.cells_end(); ++c) {
        for (int i = 0; i < 3; ++i) {
          double val = Iterate(c(), i);
          double val2 = Update(c(), i);
          if ((ValsTmp[i * 2] != -1) && ((val + val2) < ValsTmp[i * 2])) {
            Iterate(c(), i) = ValsTmp[i * 2];
          } else if ((ValsTmp[i * 2 + 1] != -1) &&
                     ((val + val2) > ValsTmp[i * 2 + 1])) {
            Iterate(c(), i) = ValsTmp[i * 2 + 1];
          } else {
            Iterate(c(), i) = val + val2;
          }
        }
      }
    }
  }
};

#endif  // FWI_FWITAPER_HPP
