#ifndef FWIFORWARDMODELDGTSVAALLATONCE_H
#define FWIFORWARDMODELDGTSVAALLATONCE_H

#include <utility>

#include "AAO/AAOIntegralPrecon.hpp"
#include "AAO/FWIIntegralOperator.hpp"
#include "AAO/FWIRKOperator.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "FWIAdjointCalculator.hpp"
#include "ForwardModel.hpp"
#include "Helper/ABCHandler.hpp"
#include "Helper/FWIDataContainerDomain.hpp"
#include "Helper/FWIDataContainerRange.hpp"
#include "Helper/GramianBlockInverse.hpp"
#include "Helper/NormCalculation.hpp"
RVector fillShapeValues(const DGAcousticElement &elem,
                        const Point &receiverPoint);

class ForwardModelAAO : public ForwardModel {

  std::shared_ptr<GramianBlockInverse> GInv = nullptr;
  double lambdaPenalty = 1.0;
  bool adaptivePenalty = false;
  double safetyFactor = 1.0;
  bool AAOVerbose = false;
  double AAOInvRegularizationValue = -1e-16;
  std::unique_ptr<Matrix> MPC = nullptr;

  void assembleAAOInv(const std::vector<Vectors> &Evals, Matrix &PHess) const;

  void getLinearizedVectors(const Vector &m, const Vectors &u_fd,
                            const Vector &Direc, std::vector<Vectors> &Evals);

  void evaluateIntegralOperatorGInverse(const Vectors &NonlinearPointWave,
                                        const Vector &nonlinearPointMat,
                                        Vectors &WaveFieldEvaluation,
                                        inteOptions opt);

  void evaluateIntegralOperator(const Vectors &NonlinearPointWave,
                                const Vector &nonlinearPointMat,
                                Vectors &WaveFieldEvaluation, inteOptions opt);

  void evaluateAndAddLinearizedIntegralOperatorGInverse(
      const Vectors &WaveFieldPoint, const Vector &MatPoint,
      const Vector &MatDirection, Vectors &WaveFieldEvaluation);

  void evaluateLinearizedIntegralOperator(const Vectors &WaveFieldPoint,
                                          const Vector &MatPoint,
                                          const Vector &MatDirection,
                                          Vectors &WaveFieldEvaluation);

  void calculateWeakFormVectors(Vectors &b) const;

  void calculateUpdateIntegral(const Vectors &WaveFieldPoint, const Vector &Mat,
                               const Vectors &Lambda, Vector &AdjointMat) const;

  void writeAdjointValuesInVector(const DGSeismogram &Seis, Vectors &U) const;
  std::shared_ptr<RKOperator> getRKOperator(const Vector &Mat, const Vectors &WaveFieldPoint,
                                            inteOptions opt);
protected:
public:
  std::shared_ptr<ABCHandlerAAO> ABCHAAO;
  ForwardModelAAO(std::shared_ptr<FWIVAcousticInversionProb> problem_fwd_,
                  std::shared_ptr<FWIDiscretizationStack> FWIDiscStackPTR_,
                  std::shared_ptr<EleCacheContainer> eleCacheContainer_,
                  std::unique_ptr<FWIObservationOperatorDGTS> Obs_,
                  std::shared_ptr<ABCHandlerAAO> ABCHAAO_,
                  std::unique_ptr<FWITaper> Taper_):
      ForwardModel(std::move(problem_fwd_),std::move(FWIDiscStackPTR_),std::move(eleCacheContainer_),
                   std::move(Obs_),std::move(Taper_)) {
    Config::Get("LambdaPenalty", lambdaPenalty);
    Config::Get("adaptivePenalty", adaptivePenalty);
    Config::Get("safetyFactor", safetyFactor);
    Config::Get("AAOVerbose", AAOVerbose);
    Config::Get("Precon", precon);
    Config::Get("AAOInvRegularizationValue", AAOInvRegularizationValue);
    GInv = std::make_shared<GramianBlockInverse>(GetWaveVec(),
                                                 *eleCacheContainer->Get());
    ABCHAAO = std::move(ABCHAAO_);
  }
  [[nodiscard]] std::string GetReconstructionType() const override { return "AllAtOnce"; }

  void AdjointMeasurement(const DGSeismogram &Seis,
                          Vectors &adjointLinearizedEvaluationWF);



  [[nodiscard]] const char *Name() const override { return "FWIForwardModelDGTSVAAllAtOnce"; }

  void NonlinearEvaluation(const FWIDataContainerDomain &nonlinearPoint,
                           FWISubproblem &evaluation,
                           std::vector<int> subproblem) override;

  void CalculateDefect(const std::vector<FWIDataContainerRange> &data_obs,
                       const FWISubproblem &nonlinearEvaluation,
                       FWISubproblem &defect,
                       std::vector<int> subproblem) override;

  void AdjointLinearized(
      const FWIDataContainerDomain &nonlinearPoint,
      const FWISubproblem &adjointLinearizedPoint,
      FWIDataContainerDomain &adjointLinearizedEvaluation) override;



  void Linearized(const FWIDataContainerDomain &nonlinearPoint,
                  const FWIDataContainerDomain &linearizedPoint,
                  FWISubproblem &linearizedEvaluation) override;

  [[nodiscard]] double ScalarProductL2(const FWIDataContainerRange &u,
                                       const FWIDataContainerRange &v) const override;

  [[nodiscard]] double ScalarProductL2(const FWIDataContainerDomain &u,
                                       const FWIDataContainerDomain &v) const override;
  [[nodiscard]] auto L2TScalarAAO(const Vectors &u, const Vectors &v) const -> double;

  [[nodiscard]] auto L2tScalarAAO(const Vector &u, const Vector &v) const -> double;





  void InitializePrecon(const Vector &m, const Vector &m_update,
                        const Vectors &u_m) override;

  void ApplyPrecon(const FWIDataContainerDomain &m,
                   FWIDataContainerDomain &m_update) override;

  void MSDProcedure(const FWIDataContainerDomain &iterateNonlinear,
                    FWIDataContainerDomain &iterateUpdatePN,
                    FWIDataContainerDomain &iterateMUpdate,
                    FWISubproblem &defectLinearized) override;

  void RestrictToVectorComp(Vector &u, int comp, int paraComps) const;

  void Linearized(const Vectors &nonlinearPointWave,
                  const Vector &nonlinearPointMat,
                  const Vectors &linearizedWave, const Vector &linearizedMat,
                  Vectors &linearizedEvaluationWave,
                  DGSeismogram &linearizedEvaluationSeismo);
  std::unique_ptr<VAcousticMatFromVec> GetMaterial
      (const FWIVAcousticInversionProb &prob, const Vector &Mat) override{
    return std::make_unique<VAcousticMatFromVec>
        (prob,Mat,nullptr,ABCHAAO.get());
  }
  void ApplyDataSpaceOp(const AAOIntegralPrecon &Prcon, const DGSeismogram &seis,
                        DGSeismogram &result, Vectors &tmp1, Vectors &tmp2);
  DGSeismogram ApplyDataSpaceCG(const AAOIntegralPrecon &Prcon, DGSeismogram &r, Vectors &tmp1,
                                Vectors &tmp2, int CGSteps);
};

#endif
