//
// Created by cr on 30/01/24.
//

#ifndef FWI_FORWARDMODELCREATOR_HPP
#define FWI_FORWARDMODELCREATOR_HPP
#include <utility>

#include "ForwardModel.hpp"
#include "ForwardModelAAO.hpp"
#include "ForwardModelRed.hpp"
#include "Observation/ObservationCreator.hpp"
class FWIForwardModelCreator {
  TimeConf TConf;
  virtual void initObs(std::shared_ptr<ObservationSpecification> spec) = 0;
protected:

  void initTaper() {
    bool taper = false;
    Config::Get("useTaper", taper);
    if (taper) {
      std::string fn;
      Config::Get("TaperFromFile", fn);
      TaperConf Conf;
      if (fn.empty()) {
        Taper = std::make_unique<FWITaper>(*DiscStack->mV,
                                           Obs->GetSpaceOp().GetSpecification(),
                                           Conf, *Problem);
      } else {
        std::string type("type");
        Config::Get("TaperFromFile", fn);
        Config::Get("TaperType", type);

        Taper = std::make_unique<FWITaper>(*DiscStack->mV, fn,type,
                                           Problem->GetParamComps());
      }
    }
  }

  void initSpecAndObs(){
    std::shared_ptr<ObservationSpecification> spec;
    std::string l;
    Config::Get("Specfile", l);
    if (l.empty()) {
      spec = ObservationSpecificationBuilder()
                 .WithT(TConf.T)
                 .Withdt(TConf.dt)
                 .Witht0(TConf.t0)
                 .CreateShared();
    } else {

      spec = std::shared_ptr<ObservationSpecification>(ObservationSpecificationBuilder().CreateFromFile());
//      int SamplingFactor = 1;
//      Config::Get("SpecfileSamplingFactor", SamplingFactor);
    }
    initObs(spec);

  }
  std::shared_ptr<FWIDiscretizationStack> DiscStack;
  std::shared_ptr<FWIVAcousticInversionProb> Problem;
  std::unique_ptr<FWIObservationOperatorDGTS> Obs;
  std::unique_ptr<FWITaper> Taper;
  std::shared_ptr<EleCacheContainer> eleCacheContainer;
  public:
  explicit FWIForwardModelCreator(std::shared_ptr<FWIVAcousticInversionProb> problem_,
                                    std::shared_ptr<FWIDiscretizationStack> disc
  ) {
    Problem =std::move(problem_);
    DiscStack = std::move(disc);
    eleCacheContainer = std::make_shared<EleCacheContainer>(*DiscStack->wV,
                                                            Problem->GetNumL());
  }

};
class ForwardModelAAOCreator : public FWIForwardModelCreator{
  bool useAAOAABC = false;
  void initObs(std::shared_ptr<ObservationSpecification> spec) override {
    Obs = ObservationCreator(eleCacheContainer->Get(),
                              DiscStack->GetMesh(),
                              Problem->GetNumL())
              .WithSpec(spec).CreateUnique();
  }

protected:
  std::shared_ptr<ABCHandlerAAO> ABCAAO;
    public:
  ForwardModelAAOCreator(std::shared_ptr<FWIVAcousticInversionProb> problem_,
                           std::shared_ptr<FWIDiscretizationStack> disc) :
          FWIForwardModelCreator(problem_,disc){
    Config::Get("UseAAOPML", useAAOAABC);
  }
  FWIForwardModelCreator& WithObs
      (std::unique_ptr<FWIObservationOperatorDGTS> Obs_){
    Obs = std::move(Obs_);
    return *this;
  }
  FWIForwardModelCreator& WithABCAAOHandler
      (std::shared_ptr<ABCHandlerAAO> ABCHandlerAAO_){
    ABCAAO = std::move(ABCHandlerAAO_);
    return *this;
  }
  ForwardModelAAO* Create(){
    if (!Obs){
      initSpecAndObs();
    }
    if(!ABCAAO) {
      if (useAAOAABC) {
        MakeABCAAO();
      }
    }
    initTaper();
    return new ForwardModelAAO(Problem,DiscStack, eleCacheContainer,
                               std::move(Obs),ABCAAO,std::move(Taper));
  }
  std::shared_ptr<ForwardModelAAO> CreateShared(){
    return std::shared_ptr<ForwardModelAAO>(Create());
  }
  void MakeABCAAO(){
      ABCConfigAAO Conf;
      Vector VEC(0.0, DiscStack->wD, DiscStack->lev);
      int ts = 0;
      Config::Get("PMLTaperStyle", ts);
      double alpha1 = Problem->GetNumL() > 0 ? Problem->alpha1() : -1;
      double alpha2 = Problem->GetNumL() > 0 ? Problem->alpha2() : -1;
      Conf.SetTaperStyle(ts, alpha1, alpha2);
      ABCAAO =
          std::make_shared<ABCHandlerAAO>(Conf, VEC, Problem->GetNumL());
  }
};
class ForwardModelRedCreator : public FWIForwardModelCreator{
  std::shared_ptr<ABCHandler> ABC;

  void initObs(std::shared_ptr<ObservationSpecification> spec) override{
    if (ABC) {
      Obs =
          ObservationCreator(ABC->GetEleCache(),
                                  ABC->GetDiscStack().GetMesh(),
                                  Problem->GetNumL())
                .WithSpec(spec)
                .CreateUnique();
    }else {
      Obs = ObservationCreator(eleCacheContainer->Get(),
                                DiscStack->GetMesh(),
                                Problem->GetNumL())
              .WithSpec(spec)
              .CreateUnique();
    }
  }

public:
  ForwardModelRedCreator(std::shared_ptr<FWIVAcousticInversionProb> problem_,
 std::shared_ptr<FWIDiscretizationStack> disc) :
FWIForwardModelCreator(std::move(problem_),std::move(disc)){}
  ForwardModelRedCreator& WithABCHandler
      (std::shared_ptr<ABCHandler> ABCHandler_){
    ABC = std::move(ABCHandler_);
    return *this;
  }
  void MakePML() {
      ABC = std::make_shared<ABCHandler>();
      ABCConf PMLConfig;
      std::vector<double> PMLPaddingLength;
      Config::Get("PMLPaddingLength", PMLPaddingLength);
      PMLConfig.SetABCPaddingLength(PMLPaddingLength);
      Config::Get("PMLBnds", PMLConfig.abcBndFlags);
      Config::Get("PMLDistFactor", PMLConfig.distfactor);
      int ts = 0;
      Config::Get("PMLTaperStyle", ts);
      double alpha1 = Problem->GetNumL() > 0 ? Problem->alpha1() : -1;
      double alpha2 = Problem->GetNumL() > 0 ? Problem->alpha2() : -1;
      PMLConfig.SetTaperStyle(ts, alpha1, alpha2);
      ABC->Initialize(DiscStack->wV->GetMesh(), PMLConfig, DiscStack->deg,
                      Problem->GetNumL(), Problem->GetParamComps());
  }
  FWIForwardModelRed* Create(){
    if(!ABC){
      bool ABCf = false;
      Config::Get("PaddingPML", ABCf);
      if(ABCf) {
        MakePML();
      }
    }
    if(!Obs){
      initSpecAndObs();
    }
    initTaper();
    return new FWIForwardModelRed(Problem,DiscStack,eleCacheContainer,
                                  ABC,std::move(Obs),std::move(Taper));
  }
  std::shared_ptr<FWIForwardModelRed> CreateShared(){
    return std::shared_ptr<FWIForwardModelRed>(Create());
  }
  std::unique_ptr<FWIForwardModelRed> CreateUnique(){
    return std::move(std::unique_ptr<FWIForwardModelRed>(Create()));
  }
};

#endif  // FWI_FORWARDMODELCREATOR_HPP
