#ifndef FWI_ABSTRACTINTERFACEDGTS
#define FWI_ABSTRACTINTERFACEDGTS

#include <iomanip>
#include <utility>

#include "AAO/FWIIntegralOperator.hpp"
#include "AAO/FWILinearizedIntegralOperator.hpp"
#include "AAO/FWIRKOperator.hpp"
#include "AlgebraFwd.hpp"
#include "Assemble.hpp"
#include "Assembler/FWILinearizedAssembler.hpp"
#include "Assembler/FWIRhsAssemblerAdjoint.hpp"
#include "Assembler/FWIRhsAssemblerLinearized.hpp"
#include "DGVAcousticAssembleBwd.hpp"
#include "DGVAcousticAssembleCombined.hpp"
#include "FWIAbstractInterfaceDGTSVA.hpp"
#include "FWIAdjointCalculator.hpp"
#include "FWITaper.hpp"
#include "ForwardModel.hpp"
#include "Helper/FWIDataContainerDomain.hpp"
#include "Helper/FWIDataContainerRange.hpp"
#include "Helper/FWIHelpers.hpp"
#include "Helper/FWIMppHelpers.hpp"
#include "Helper/GramianBlockInverse.hpp"
#include "Helper/NormCalculation.hpp"
#include "Helper/VectorsIntegrator.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "PseudoHessianPreconditioner.hpp"
#include "Reduced/FWILinOpAdjRed.hpp"
#include "Reduced/FWILinOpRed.hpp"
#include "TimeIntegrator.hpp"
#include "VAcousticMatFromVec.hpp"
#include "lib1_math/basicalgebra/RMatrix.hpp"
#include "lib4_fem/elements/Element.hpp"

#include "m++.hpp"

class FWIForwardModelRed : public ForwardModel {
  std::unique_ptr<FWILinOpAdjRed> AdjOp;
  std::unique_ptr<FWILinOpRed> LinOp;
  std::unique_ptr<PseudoHessianPreconditioner> PHess;

protected:
public:
  FWIForwardModelRed(std::shared_ptr<FWIVAcousticInversionProb> problem_fwd_,
                     std::shared_ptr<FWIDiscretizationStack> FWIDiscStackPTR_,
                     std::shared_ptr<EleCacheContainer> eleCacheContainer_,
                     const std::shared_ptr<ABCHandler> &PMLH_,
                     std::unique_ptr<FWIObservationOperatorDGTS> Obs_,
                     std::unique_ptr<FWITaper> Taper_)
      : ForwardModel(std::move(problem_fwd_),std::move(FWIDiscStackPTR_),std::move(eleCacheContainer_),
                     std::move(Obs_),std::move(Taper_))
  {
    ABCH = PMLH_;
    Config::Get("FWIPrecon", precon);
  }
  virtual void MakeLinearAndAdjointOperator(const Vector& m, Vectors& uM){
    mout << "Assembling Linearized Operators" << endl;
    AdjOp = std::make_unique<FWILinOpAdjRed>(uM, m, *Obs, problem_fwd, ABCH, TConf, recon,
                                             eleCacheContainer->Get());
    LinOp = std::make_unique<FWILinOpRed>(uM, m, problem_fwd, ABCH, TConf, eleCacheContainer->Get(),
                                          false);
    mout << "Done!" << endl;
  }
  ~FWIForwardModelRed() override;
  void NonlinearEvaluation(const FWIDataContainerDomain &nonlinearPoint,
                           FWISubproblem &evaluation,
                           std::vector<int> subproblem) override;
  virtual void Nonlinear(const Vector& m, Vectors& uM, DGSeismogram &sM,int shot){
    Parameter2SeismogramMap(m,sM,shot,&uM);
  }

  void InitializePrecon(const Vector &M, const Vector &M_update,
                        const Vectors &U_M) override;
  const Mesh &GetMesh() {
    if (ABCH) return ABCH->GetDiscStack().meshes->fine();
    return (*FWIDiscStack.meshes)[FWIDiscStack.lev];
  }
  [[nodiscard]] const char *Name() const override { return "FWDModelVAcoustic"; }
  void AdjointLinearized(const FWIDataContainerDomain &nonlinearPoint,
                         const FWISubproblem &adjointLinearizedPoint,
                         FWIDataContainerDomain &adjointLinearizedEvaluation) override;

  void Linearized(const FWIDataContainerDomain &nonlinearPoint,
                  const FWIDataContainerDomain &linearizedPoint,
                  FWISubproblem &linearizedEvaluation) override;
  void CalculateDefect(const std::vector<FWIDataContainerRange> &data_obs,
                       const FWISubproblem &nonlinearEvaluation,
                       FWISubproblem &defect,
                       std::vector<int> subproblem) override;
  std::unique_ptr<VAcousticMatFromVec>
      GetMaterial(const FWIVAcousticInversionProb &prob, const Vector &Mat) override;
  [[nodiscard]] double ScalarProductL2(const FWIDataContainerRange &u,
                                       const FWIDataContainerRange &v) const override {
    const auto &s1 = u.getSeismogram();
    const auto &s2 = v.getSeismogram();
    return s1 * s2;
  }
  [[nodiscard]] double ScalarProductL2(const FWIDataContainerDomain &u,
                                       const FWIDataContainerDomain &v) const override {
    const auto &UU(u.getMat());
    const FWIDataContainerDomain &vv(v);
    const auto &VV(vv.getMat());
    return L2Scalar(UU, VV);
  }
  void ApplyPrecon(const FWIDataContainerDomain &m,
                   FWIDataContainerDomain &m_update) override {
    if(precon) PHess->Apply(m_update.getMatWrite());
  }

  bool UseGardener() override { return useGardener; }

  [[nodiscard]] int GetDim() const;

  [[nodiscard]] double L2TScalar(const Vectors &u, const Vectors &v) const;

  [[nodiscard]] double L2tScalar(const Vector &u, const Vector &v) const;

  [[nodiscard]] double L2Scalar(const Vector &u, const Vector &v) const;
};

#endif
