//
// Created by cr on 21.11.22.
//

#include "VTUModelReader.hpp"

void VTUModelReader::checkLines(const std::string &line_s, std::vector<std::ifstream> &fileStreams, size_t nrOfFiles) {
    if (nrOfFiles == 1) { return; }
    std::vector<bool> checkOk(nrOfFiles - 1, false);
    std::string line_tmp = "";
    for (size_t i = 0; i < nrOfFiles - 1; ++i) {
        getline(fileStreams[i + 1], line_tmp);
        if (line_tmp == line_s) {
            checkOk[i] = true;
        }
    }
    for (size_t i = 0; i < nrOfFiles - 1; ++i) {
        if (!checkOk[i]) mout << "The file nr " << i << " differs from the reference" << endl;
    }
    for (size_t i = 0; i < nrOfFiles - 1; ++i) {
        if (!checkOk[i]) Exit("error detected!");
    }
}

void VTUModelReader::createLookuplistWithoutSave(const VectorMatrixBase &mgr) {
    size_t nrOfFiles = filenames.size();
    std::vector<std::ifstream> fileStreams(nrOfFiles);
    for (size_t i = 0; i < nrOfFiles; ++i) {
        fileStreams[i].open(filenames[i].c_str());
        if (fileStreams[i].fail()) {
            Exit("Could not open the file: " + filenames[i]);
        } else {
            mout << "File " << filenames[i] << " opened!" << endl;
        }
    }
    std::string line_s;
    std::string line_tmp = "";
    size_t nrcells = 0;
    for (cell c = mgr.cells(); c != mgr.cells_end(); ++c) {
        ++nrcells;
    }
    std::vector<bool> checkOk(true, nrOfFiles);
    bool POINTS = false;
    bool CELLS = false;
    bool BREAKFLAG = true;
    std::unordered_map<int, Point> points_vec;
    while (!fileStreams[0].eof() && BREAKFLAG) {
        getline(fileStreams[0], line_s);
        checkLines(line_s, fileStreams, nrOfFiles);
//        MOUT(line_s)
        if (line_s.rfind("<Points>") < 50) {
            POINTS = true;
            MOUT("points true!")
            std::stringstream line(line_s);
            int size;
            std::string s;
            line >> s;
            line >> size;
            getline(fileStreams[0], line_s);
            checkLines(line_s, fileStreams, nrOfFiles);
            continue;
        }
        int cnt = 0;
        while (POINTS) {
            std::stringstream line(line_s);
            std::vector<double> values;
            double d;
            int i = 0;
            Point p(0.0);
            while (line >> d) {
                p[i] = d;
                i = (i + 1) % 3;
                if (i == 0) {
                    points_vec[cnt] = p;
//                    MOUT(p)
//                    MOUT(cnt)
                    cnt++;
                }
            }
            if (i != 0) Exit("Point Dimension does not match!")
            getline(fileStreams[0], line_s);
            checkLines(line_s, fileStreams, nrOfFiles);
            if (line_s.rfind("</DataArray>") < 50) {
                POINTS = false;
                while (!CELLS) {
                    getline(fileStreams[0], line_s);
                    if (line_s.rfind("<Cells>") < 50) CELLS = true;
                }
                getline(fileStreams[0], line_s);
                continue;
            }
        }
        int lookup = 0;
//        getline(fileStreams[0],line_s);
        checkLines(line_s, fileStreams, nrOfFiles);
        while (CELLS) {
            double pp1 = 0.0, pp2 = 0.0;
            std::stringstream line(line_s);
//            MOUT(line_s);
            int i = 0;
            int corners = 4;
//            if (i != 4){
//                Exit("only for squares implemented");
//            }
            int count = 0;
            while (line >> i) {
                count = (count + 1) % corners;
                const Point &p = points_vec[i];
//                MOUT(points_vec[i]);
                pp1 += p[0] / corners;
                pp2 += p[1] / corners;
                if (count == 0) {
//                    MOUT(pp1)
//                    MOUT(pp2)
                    cell *c = foundPoint(mgr, pp1, pp2);
                    if (c != nullptr) {
                        if (lookupMap.find((*c)()) == lookupMap.end()) {
                            lookupMap[(*c)()] = lookup;
                        }
                    }
                    pp1 = 0.0;
                    pp2 = 0.0;
                    lookup++;
                }
            }
            getline(fileStreams[0], line_s);
            checkLines(line_s, fileStreams, nrOfFiles);
            if (line_s.rfind("</DataArray>") < 50) {
                CELLS = false;
                BREAKFLAG = false;
            }
        }
    }
//    MOUT(nrcells)
//    MOUT(lookupMap.size())
    if (nrcells != lookupMap.size()) {Exit("number of cells of VTK and mesh are not the same!"); }
}

void VTUModelReader::readPointData(std::vector<double> &dataVTK,
                                   const std::string &filenameVTU,
                                   const std::string &dataStr) const {
  std::ifstream fileStreamVTU;
  fileStreamVTU.open(filenameVTU.c_str());
  if (fileStreamVTU.fail()) {
    Exit("Could not read VTK with the filename: " + filenameVTU);
  }
  std::string line_s = "";
  bool LOOKUP = false;
  while (!fileStreamVTU.eof()) {
    getline(fileStreamVTU, line_s);
    if (LOOKUP) {
      std::stringstream line(line_s);
      double d;
      while (line >> d) {
        dataVTK.push_back(d);
      }
    }
    if (line_s.rfind("CellData Scalars=\"" + dataStr + "\"") < 50) {
      LOOKUP = true;
      getline(fileStreamVTU, line_s);
    }
    if (line_s.rfind("/DataArray") < 50 && LOOKUP) {
      break;
    }
  }
}

void VTUModelReader::getData(Vector &M, std::string &DataStr, int param) {
    std::vector<double> dataVTK;
    readPointData(dataVTK, filenames[currentIterate], DataStr);
    for (cell c = M.cells(); c != M.cells_end(); ++c) {
        M(c(), param) = dataVTK[lookupMap[c()]];
    }
}

cell *VTUModelReader::foundPoint(const VectorMatrixBase &mgr, double pp1, double pp2) {
    for (cell c = mgr.cells(); c != mgr.cells_end(); ++c) {
        double tmp = pp1 - c()[0];
        double tmp2 = pp2 - c()[1];
        double d = sqrt(tmp * tmp + tmp2 * tmp2);
        if (d < 1e-1) {
            return new cell(c);
        }
    }
    return nullptr;
}