//
// Created by cr on 19.11.22.
//

#include "VectorFillFunctions.hpp"

void ReparametrizeMaterialVector(const Vector &Mat, const FWIVAcousticInversionProb &prob, Vector &MatRepara) {
    MatRepara.Clear();
    int numL = prob.GetNumL();
    if (prob.getParametrization() == "native") {
        MatRepara = Mat;
    } else if (prob.getParametrization() == "taup") {
        for (cell c = MatRepara.cells(); c != MatRepara.cells_end(); ++c) {
            double V_p = Mat(c(), 1);
            double Rho =  Mat(c(), 0);
            double Tau_p = (numL > 0) ? Mat(c(), 2) : 0;
            double alpha = (numL > 0) ? prob.alpha(*c, c()) : 0;
            MatRepara(c(), 0) = Rho;
            MatRepara(c(), 1) = (V_p * V_p * Rho) / (1 + alpha * Tau_p);
            if (numL > 0) {
                for (int i = 0; i < numL; ++i) {
                    MatRepara(c(), 2 + i) = (V_p * V_p * Rho * Tau_p) / (1 + alpha * Tau_p);
                }
            }

        }
    } else if (prob.getParametrization() == "taup_positive") {
        for (cell c = MatRepara.cells(); c != MatRepara.cells_end(); ++c) {
            double V_p = exp(Mat(c(), 1));
            double Rho =  exp(Mat(c(), 0));
            double Tau_p = (numL > 0) ? exp(Mat(c(), 2)) : 0;
            double alpha = (numL > 0) ? prob.alpha(*c, c()) : 0;
            MatRepara(c(), 0) = Rho;
            MatRepara(c(), 1) = (V_p * V_p * Rho) / (1 + alpha * Tau_p);
            if (numL > 0) {
                for (int i = 0; i < numL; ++i) {
                    MatRepara(c(), 2 + i) = (V_p * V_p * Rho * Tau_p) / (1 + alpha * Tau_p);
                }
            }
        }
    } else {
        Exit("Unknown parameterization");
    }
}

void FillVectorFromProblem(const FWIVAcousticInversionProb &prob, Vector &M) {
    M.Clear();
    for (cell c = M.cells(); c != M.cells_end(); ++c) {
        FillCellFromProblem(c, prob, M);
    }
}

void FillCellFromProblem(const cell &c, const FWIVAcousticInversionProb &prob, Vector &M) {
    int numL = prob.GetNumL();
    if (prob.getParametrization() == "native") {
        M(c(), 0) = prob.Rho(*c, c());
        M(c(), 1) = prob.Kappa_i(*c, c(), 0);
        for (int i = 1; i < numL + 1; ++i) {
            M(c(), 1 + i) = prob.Kappa_i(*c, c(), i);
            M(c(), numL + i + 1) = prob.Tau_i(*c, c(), i);
        }
    } else if (prob.getParametrization() == "taup") {
        M(c(), 0) = prob.Rho(*c, c());
        M(c(), 1) = prob.V_p(*c, c());
        if (numL > 0) {
            M(c(), 2) = prob.Tau_p(*c, c());
            for (int i = 1; i <= numL; ++i) {
                M(c(), 2 + i) = prob.Tau_i(*c, c(), i);
            }
        }
    } else if (prob.getParametrization() == "taup_positive") {
        M(c(), 0) = log(prob.Rho(*c, c()));
        M(c(), 1) = log(prob.V_p(*c, c()));
        if (numL > 0) {
            M(c(), 2) = log(prob.Tau_p(*c, c()));
            for (int i = 1; i <= numL; ++i) {
                M(c(), 2 + i) = prob.Tau_i(*c, c(), i);
            }
        }
    } else {
        Exit("unknown parameterization");
    }
}
