//
// Created by cr on 07/02/24.
//

#include <fftw3.h>

#include "DGSeismogram.hpp"
#include "FFT.hpp"
//#include "IFOSToFWI.hpp"
#include "../../../DSPFilters/Butterworth.h"
void DGSeismogram::BandPassFilter(double lowerFreq, double upperFreq, int order) {
//  THROW("Update to new Datastructure!")

  if (SeismoData.GetParaDat().spec.GetNrOfMeasure() > 1) {
    THROW("Only for one state component defined!");
  }
  double centerFreq = (upperFreq + lowerFreq)/ 2.0;
  double width = upperFreq - centerFreq;
  int numSamples = SeismoData.GetParaDat().spec.GetNumberOfTimeSteps();
  double *seismoData[1];
  seismoData[0] = new double[numSamples];
  double *data[1];
  data[0] = new double[numSamples];
  for (const auto &re : SeismoData.GetParaDat().LocalReceivers) {
      auto  timeSeries = GetTimeSeries(re);
      auto f = std::make_unique<
          Dsp::SimpleFilter<Dsp::Butterworth::BandPass<4>, 1>>();
      f->setup(order, 1.0 / SeismoData.GetParaDat().spec.GetTimeDelta(), centerFreq, width);
      for (int i = 0; i < timeSeries.size(); ++i) {
        seismoData[0][i] = timeSeries[i];
      }
      f->process(numSamples, seismoData);
      for (int i = 0; i < timeSeries.size(); ++i) {
        data[0][i] = seismoData[0][timeSeries.size() - 1 - i];
      }
      f->process(numSamples, data);
      for (int i = 0; i < timeSeries.size(); ++i) {
        timeSeries[timeSeries.size() - 1 - i] = data[0][i];
      }
  }
  delete[] data[0];
  delete[] seismoData[0];
}
void DGSeismogram::BandPassFilterIFOS(double lowerFreq, double upperFreq) {
  double F_HP = lowerFreq;
  double fc = upperFreq;
  int NT = SeismoData.GetParaDat().spec.GetNumberOfTimeSteps();
  int Npad = (int)(pow(2.0, ceil(log((double)(NT))/log(2.0))+1.0) );
  double df = (double) 1.0/(Npad)/ SeismoData.GetParaDat().spec.GetTimeDelta();
  std::vector<double> ff(Npad + 1), BFabs(Npad + 1);
  for (int j=0;j<Npad/2;j++){
    ff[j+1] = (double) j * df;
  }
  for (int j=Npad/2;j<Npad;j++){
    ff[j+1] = -(double) (Npad-j) * df;
  }
  int it1=floor(1/df);
  int it2=round(F_HP/df);
  int it3=floor(fc/df);
  int it4 = 2*it3;
  int flank1 = it2 - it1;
  int flank2 = it4-it3;
  if(PPM->Proc()==0) {
    printf("    Band-pass corner indexes    : it1 = %d, it2 = %d, it3 = %d, it4 = %d \n",
           it1,it2,it3,it4);
    printf("    Band-pass corner frequencies: Fc1 = %.2f, Fc2 = %.2f, Fc3 = %.2f, Fc4 = %.2f \n",
           ff[it1],ff[it2],ff[it3],ff[it4]);
  }
  BFabs[0] = 0.0;
  for (int n=0; n<=Npad; n++) {
    if ((n>=it1) && (n<it3))	BFabs[n] = 1.0;
    if ((n>=it3) && (n<it4))	BFabs[n] = (double) (1.0-(1-cos((double)(n-it3)/flank2*M_PI))/2.0);
    if(n>=it4) BFabs[n] = 0.0; // BFabs[n];
    if(F_HP) {
      if (n<=it1) BFabs[n] = 0.0;
      if ((n>it1) && (n<it2))	BFabs[n] = (double) ((1.0-cos((double)(n-it1)/flank1*M_PI))/2.0);
      if ((n>=it2) && (n<it3))	BFabs[n] = 1.0;
    }
  }
  for (const auto& re : SeismoData.GetParaDat().LocalReceivers){
    std::vector<std::complex<double>> in_data(Npad,0.0);
    std::vector<std::complex<double>> out_data(Npad,0.0);

    auto timeSeries = GetTimeSeries(re);
    if (timeSeries.empty()) continue;
    for (int i = 0; i < timeSeries.size()-1; ++i) {
      in_data[i].real(timeSeries[i+1]);
      in_data[i].imag(0.0);
      out_data[i].real(0.0);
      out_data[i].imag(0.0);
    }
    for (int j = timeSeries.size()-1; j < Npad; j++) {
      in_data[j].real(0.0);
      in_data[j].imag(0.0);
      out_data[j].real(0.0);
      out_data[j].imag(0.0);
    }
    fftw_plan p_data;
    p_data = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(in_data.data()),
                              reinterpret_cast<fftw_complex *>(out_data.data()), 1, FFTW_ESTIMATE);
    fftw_execute(p_data);
    for (int j = 1; j <= Npad / 2; j++) {
      out_data[j] = out_data[j] * BFabs[j];
      if( j!= 1) {
        out_data[Npad + 1 - j] = conj(out_data[j]) * BFabs[j];
      }
      in_data[j].imag(0.0);
      in_data[j].real(0.0);
    }
    out_data[0].real(0.0);
    out_data[0].imag(0.0);
    out_data[1].real(0.0);
    out_data[1].imag(0.0);
    fftw_destroy_plan(p_data);
    fftw_plan p_data2;

    p_data2 = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(out_data.data()),
                               reinterpret_cast<fftw_complex *>(in_data.data()), -1, FFTW_ESTIMATE);
    fftw_execute(p_data2);
    for (int i = 1; i < timeSeries.size(); ++i) {
      timeSeries[i] = in_data[i-1].real()/Npad;
    }
    fftw_destroy_plan(p_data2);
  }
}
void DGSeismogram::fillSeismogramWithNormalGaussianNoise(
    DGSeismogram &noiseSeismo, unsigned int seed) {

  noiseSeismo =0.0;
  std::random_device rd;
  unsigned int localSeed = seed == 0 ? rd() : seed;
  unsigned int rnd = localSeed+PPM->Proc()*1000;
  std::normal_distribution<double> distribution(0.0, 1.0);
  for (const auto &re : noiseSeismo.SeismoData.GetParaDat().LocalReceivers) {
    std::mt19937 generator(rnd++);
    for (auto &t : noiseSeismo.GetTimeSeries(re)) {
      t = distribution(generator);
    }
  }
  double noisePower = noiseSeismo*noiseSeismo;
  noiseSeismo *= 1.0/sqrt(noisePower);
}
void DGSeismogram::addRelativeNoise(double signalToNoiseRatio, unsigned int seed) {
  auto noiseSeismo = DGSeismogram(*this);
  fillSeismogramWithNormalGaussianNoise(noiseSeismo, seed);
  double signalPower = sqrt((*this)*(*this));
  noiseSeismo *=  signalPower/ signalToNoiseRatio;
  (*this) += noiseSeismo;
}
CMatrix DGSeismogram::GetFFT() const {

  if (SeismoData.GetParaDat().spec.GetNrOfMeasure() > 1) {
    THROW("Only for one state component defined!");
  }
  CMatrix FFTMat(0.0,0,0);
  RVector SeismoDataInput(0.0,
                          SeismoData.GetParaDat().spec.GetNumberOfTimeSteps());
  CVector SeismoDataOutput(0.0,
                           SeismoData.GetParaDat().spec.GetNumberOfTimeSteps());
  int j = 0;
  for (const auto& re : SeismoData.GetParaDat().LocalReceivers){
    auto timeSeries = GetTimeSeries(re);
    std::copy(timeSeries.begin(),timeSeries.end(),SeismoDataInput.begin());
    FFT::RealToComplexVector(SeismoDataInput,SeismoDataOutput);
    for (int i = 0; i < timeSeries.size(); ++i) {
      FFTMat[j][i] = SeismoDataOutput[i];
    }
    ++j;
    SeismoDataOutput = 0.0;
  }
  return FFTMat;
}

void DGSeismogram::LowPassFilter(double freq, int order) {
    if (SeismoData.GetParaDat().spec.GetNrOfMeasure() > 1) {
    THROW("Only for one state component defined!");
  }
  int numSamples = SeismoData.GetParaDat().spec.GetNumberOfTimeSteps();
  double *seismoData[1];
  seismoData[0] = new double[numSamples];
  double *data[1];
  data[0] = new double[numSamples];
  for (const auto &re : SeismoData.GetParaDat().LocalReceivers) {
    auto f = std::make_unique<Dsp::SimpleFilter<Dsp::Butterworth::LowPass<4>, 1>>();
    f->setup(order, 1.0 / SeismoData.GetParaDat().spec.GetTimeDelta(), freq);
    auto  timeSeries = GetTimeSeries(re);
    for (int i = 0; i < timeSeries.size(); ++i) {
      seismoData[0][i] = timeSeries[0];
    }
    f->process(numSamples, seismoData);
    for (int i = 0; i < timeSeries.size(); ++i) {
      data[0][i] = seismoData[0][timeSeries.size() - 1 - i];
    }
    f->process(numSamples, data);
    for (int i = 0; i < timeSeries.size(); ++i) {
      timeSeries[timeSeries.size() - 1 - i] = data[0][i];
    }
  }
  delete[] data[0];
  delete[] seismoData[0];
}
double operator*(const DGSeismogram &s1, const DGSeismogram &s2) {
  double res2{};
  if (s1.SeismoData.GetData().size() > 0) {
    res2 = s1.SeismoData.GetData() * s2.SeismoData.GetData();
    for (size_t i = 0; i < s1.SeismoData.ReceiverSize(); ++i) {
      res2 -= 0.5 * (s1.SeismoData.GetData()[s1.SeismoData.GetLocToRange(i)[0]]
                     *s2.SeismoData.GetData()[s2.SeismoData.GetLocToRange(i)[0]]);
      res2 -= 0.5 * (s1.SeismoData.GetData()[s1.SeismoData.GetLocToRange(i)[1]]
                     *s2.SeismoData.GetData()[s2.SeismoData.GetLocToRange(i)[1]]);
    }
  }
  return PPM->Sum(res2) * s1.SeismoData.GetParaDat().spec.GetTimeDelta();
}