//
// Created by cr on 07/02/24.
//

#ifndef FWI_DGSEISMOGRAM_HPP
#define FWI_DGSEISMOGRAM_HPP
#include <string>
#include "AlgebraFwd.hpp"
#include "lib6_app/seismogram/SeismogramData.hpp"
class DGSeismogram {
  SeismogramData SeismoData;
public:
  DGSeismogram() = delete;
  explicit DGSeismogram(const SeismogramParallelizationData &ParaDat_)
      : SeismoData(ParaDat_){}
  DGSeismogram(const DGSeismogram& ) = default;
  const SeismogramData &GetData() const { return SeismoData;
  }
  SeismogramData &GetData() { return SeismoData; }
  void WriteToFile(const std::string& fn) const {
     SeismoData.WriteToFile(fn);
   }
   void ReadFromFile(const std::string& fn) {
     SeismoData.ReadFromFile(fn);
   }
  void LowPassFilter(double, int);
  void BandPassFilter(double lowerFreq, double UpperFreq, int order);
  void BandPassFilterIFOS(double lowerFreq, double upperFreq);
  void fillSeismogramWithNormalGaussianNoise(DGSeismogram &noiseSeismo, unsigned int seed);
  void addRelativeNoise(double signalToNoiseRatio, unsigned int seed);
  CMatrix GetFFT() const;
  auto GetTimeSeries(const Point &receiver) const  {
    return SeismoData.GetTimeseries(receiver);
  }
  auto GetTimeSeries(const Point &receiver)   {
    return SeismoData.GetTimeseries(receiver);
  }
  const ObservationSpecification &getSpecification() const{
    return SeismoData.GetParaDat().spec;
  }

  DGSeismogram& operator=(const DGSeismogram& s){
    SeismoData = s.SeismoData;
    return *this;
  }
  virtual void operator=(double val){
    SeismoData =val;
  };

  void operator*=(double val){
    SeismoData *= val;
  }

  void operator-=(const DGSeismogram &s){
    SeismoData -= s.SeismoData;
  }

  void operator+=(const DGSeismogram &s) { SeismoData += s.SeismoData; }

  friend double operator*(const DGSeismogram &s1, const DGSeismogram &s2);

  DGSeismogram &operator+=(const constAB<double, DGSeismogram> &aS) {
    double s = aS.first();
    const auto &seisData = aS.second().GetData().GetData();
    GetData().GetDataWrite() += s * seisData;
    return *this;
  }
  DGSeismogram &operator-=(const constAB<double, DGSeismogram> &aS) {
    double s = aS.first();
    const auto &seisData = aS.second().GetData().GetData();
    GetData().GetDataWrite() -= s * seisData;
    return *this;
  }
};
inline constAB<double, DGSeismogram> operator*(double &s,
                                               const DGSeismogram &S) {
  return constAB<double, DGSeismogram>(s, S);
}

#endif  // FWI_DGSEISMOGRAM_HPP
