//
// Created by cr on 25/01/24.
//

#ifndef FWI_FWIOBSERVATIONOPERATORSPACEDGTS_HPP
#define FWI_FWIOBSERVATIONOPERATORSPACEDGTS_HPP

#include <memory>
#include <unordered_set>
#include <utility>

#include "DGSeismogram.hpp"
#include "Helper/FWIHelpers.hpp"
#include "Helper/FWIMppHelpers.hpp"
#include "MeasureMask.hpp"
#include "VectorField.hpp"
#include "lib6_app/seismogram/ObservationSpecification.hpp"

class FWIObservationOperatorSpaceDGTS {
protected:
    std::shared_ptr<ObservationSpecification> spec;
    std::shared_ptr<EleCacheType> EleCache;
    int measureComp = 0;
    int stateComps = -1;
    mutable std::unordered_map<Point, std::unordered_map<Point, std::vector<VectorField>>> VFVals;
    mutable std::unordered_map<Point, std::unordered_map<Point, RVector>> PVals;
    mutable std::unordered_map<Point, std::__cxx11::list<Point>> SpaceReceiver2SpaceSupport;
    std::unique_ptr<SeismogramParallelizationData> ParaDat;
public:
    FWIObservationOperatorSpaceDGTS() = delete;

    explicit FWIObservationOperatorSpaceDGTS(std::shared_ptr<EleCacheType> EleCache_,
                                             std::shared_ptr<ObservationSpecification> spec_) : spec(std::move(spec_)),
                                                                                               EleCache(std::move(EleCache_)),
                                                                                               stateComps(
                                                                                                       EleCache->begin()->second->getDim() +
                                                                                                       EleCache->begin()->second->p_dimension())
           {
    }

    virtual void SetSpaceValue(const FWIState &Measure, const Point &Receiver,
                               const std::list<Point> &MeasureCells,
                               Vector &b) const = 0;

    const std::unordered_map<Point, std::list<Point>> &GetSpaceSupport() const {
      return SpaceReceiver2SpaceSupport;
    }
    const SeismogramParallelizationData &GetParaDat() const {
      return *ParaDat;
    }

    bool MeasureHere(const Point &receiver) const {
      return SpaceReceiver2SpaceSupport.contains(receiver);
    }

    virtual FWIState GetSpaceValue(const Vector &, const Point &, const std::list<Point> &) const = 0;

    virtual ~FWIObservationOperatorSpaceDGTS() = default;

    const ObservationSpecification &GetSpecification() const {return *spec;}

  virtual std::string Name() const = 0;
};

class FWIObservationOperatorSpacePoint : public FWIObservationOperatorSpaceDGTS {
    std::unordered_map<Point, unsigned> ReceiverToCellCount;

    void initializeReceiverToCellCount(const ObservationSpecification &spec) {
        for (const auto &receiver: spec) {
            const unsigned cell_count = PPM->Sum(GetSpaceSupport().contains(receiver) ?
                                                 GetSpaceSupport().at(receiver).size() : 0);
            ReceiverToCellCount[receiver] = cell_count;
        }
    }

    void initalizeShapeValues(const ObservationSpecification &spec, const Mesh &b);
void initializeParallelizationData(const ObservationSpecification &spec){
  static constexpr double Eps = 1e-10;

  ParaDat = std::make_unique<SeismogramParallelizationData>(spec);
    for (const auto& re : SpaceReceiver2SpaceSupport){
      ParaDat->LocalReceivers.emplace_back(re.first);
    }
    std::sort(ParaDat->LocalReceivers.begin(),ParaDat->LocalReceivers.end());
    ParaDat->LocalIndexToGlobalIndex =
        std::vector<int>(ParaDat->LocalReceivers.size());
    for (int i =0 ; i<ParaDat->LocalReceivers.size();++i){
      int j = 0;
      for(const auto & re : spec){
        if (norm((re -ParaDat->LocalReceivers[i])) < Eps){
          ParaDat->CoordsToLocalIndex[re] = i;
          ParaDat->LocalIndexToGlobalIndex[i] =j;
        }
        ++j;
      }
    }
    for (int i = 0; i < ParaDat->CoordsToLocalIndex.size();++i){
      ParaDat->LocalIndexToLocalRange.push_back({(i * spec.GetNumberOfTimeSteps()),((i+1) * spec.GetNumberOfTimeSteps()-1)});

      ParaDat->LocalIndexToGlobalRange.push_back({ (ParaDat->LocalIndexToGlobalIndex[i] * spec.GetNumberOfTimeSteps()),
                                              ((ParaDat->LocalIndexToGlobalIndex[i]+1) *
                spec.GetNumberOfTimeSteps()-1)});
      ParaDat->GlobalTimesOwned.push_back({0, spec.GetNumberOfTimeSteps()-1});
    }
  }

public:
    FWIObservationOperatorSpacePoint() = delete;

    FWIObservationOperatorSpacePoint(std::shared_ptr<EleCacheType> EleCache,
                                     std::shared_ptr<ObservationSpecification> spec_,
                                     const Mesh &mesh, int MeasureComp)
            : FWIObservationOperatorSpaceDGTS(std::move(EleCache), std::move(spec_)) {
        measureComp = MeasureComp;
        initalizeShapeValues(*spec, mesh);
        initializeReceiverToCellCount(*spec);
        initializeParallelizationData(*spec);
    }

    std::string Name() const override { return "point"; }
    void SetSpaceValue(const FWIState &Measure, const Point &Receiver,
                       const std::list<Point> &MeasureCells,
                       Vector &b) const override;

    FWIState GetSpaceValue(const Vector &U, const Point &Receiver, const std::list<Point> &MeasureCells) const override;

};




#endif //FWI_FWIOBSERVATIONOPERATORSPACEDGTS_HPP
