//
// Created by cr on 15.11.22.
//

#ifndef FWI_MEASUREMASK_HPP
#define FWI_MEASUREMASK_HPP

#include <memory>
#include "RVector.hpp"

using FWIState = RVector;

class MeasureMask {
public:
    MeasureMask() = default;

    virtual void mask(FWIState &) const {Exit("not implemented"); }

    virtual void maskAssemble(FWIState &) const {Exit("not implemented"); }

    virtual void adjoint(FWIState &) const {Exit("not implemented"); }

    virtual ~MeasureMask() = default;

    [[nodiscard]] virtual int GetSeismoComps() const {Exit("not implemented"); }

    [[nodiscard]] virtual int GetStateComps() const {Exit("not implemented"); }
};

class MaskEverything : public MeasureMask {
    int Comps;
public:
    MaskEverything(int _numL, int _dim) : Comps(1 + _numL + _dim) {}

    void mask(FWIState &) const override {}

    void adjoint(FWIState &) const override {}

    [[nodiscard]] int GetSeismoComps() const override { return Comps; }

    [[nodiscard]] int GetStateComps() const override { return Comps; }

    ~MaskEverything() override = default;
};

class MaskPressureSumVA : public MeasureMask {
    int seismoComp = 1;
    int numL;
    int dim;
public:
    MaskPressureSumVA(int _numL, int _dim) : numL(_numL), dim(_dim) {}

    void mask(FWIState &v) const override;

    void adjoint(FWIState &v) const override;

    void maskAssemble(FWIState &) const override;

    [[nodiscard]] int GetSeismoComps() const override { return seismoComp; }

    [[nodiscard]] int GetStateComps() const override { return dim + numL + 1; }
};

MeasureMask *GetMeasureMask(int dim);

#endif //FWI_MEASUREMASK_HPP
