//
// Created by cr on 11.11.22.
//

#include <utility>

#include "FWIObservationOperatorSpaceDGTS.hpp"
#include "FWIObservationOperatorTimeDGTS.hpp"
#include "ObservationCreator.hpp"

auto ObservationCreator::CreateShared() -> std::shared_ptr<FWIObservationOperatorDGTS> {
    return std::shared_ptr<FWIObservationOperatorDGTS>(Create());
}

auto ObservationCreator::CreateUnique() -> std::unique_ptr<FWIObservationOperatorDGTS> {
    return std::unique_ptr<FWIObservationOperatorDGTS>(Create());
}
void ObservationCreator::getSpec() {
  spec = ObservationSpecificationBuilder().CreateShared();
}

FWIObservationOperatorDGTS *ObservationCreator::Create() {
  if (!spec) {
    getSpec();
  }
    if (!Mask) {
        Mask = GetMeasureMask(mesh.dim(), numL, MeasureComp);
    }

    spaceOp = std::make_unique<FWIObservationOperatorSpacePoint>(eleCache, spec, mesh,
                                                                 MeasureComp); // SpacePoint Hardcoded
    timeOp = std::make_unique<FWIObservationOperatorTimePointDGTS>(spec);                        // TimePoint Hardcoded
    return new FWIObservationOperatorDGTS(std::move(timeOp), std::move(spaceOp), std::move(Mask));
}

ObservationCreator &ObservationCreator::WithSpec(std::shared_ptr<ObservationSpecification> spec_) {
    spec = std::move(spec_);
    return *this;
}

ObservationCreator &ObservationCreator::WithMask(std::unique_ptr<MeasureMask> Mask_) {
    Mask = std::move(Mask_);
    return *this;
}

ObservationCreator &ObservationCreator::WithSpaceOp(std::unique_ptr<FWIObservationOperatorSpaceDGTS> spaceOp_) {
    spaceOp = std::move(spaceOp_);
    return *this;
}

ObservationCreator &ObservationCreator::WithTimeOp(std::unique_ptr<FWIObservationOperatorTimeDGTS> timeOp_) {
    timeOp = std::move(timeOp_);
    return *this;
}

ObservationCreator::ObservationCreator(std::shared_ptr<EleCacheType> eleCache_,
                                             const Mesh &mesh_, int numL_) :
        eleCache(std::move(eleCache_)), mesh(mesh_) {
    numL = numL_;
    readConfig();
}

void ObservationCreator::readConfig() {
    Config::Get("MeasureComp", MeasureComp);
    if (!(MeasureComp == 0 || MeasureComp == 1)) THROW("Measurement Configuration unknown");
}

