//
// Created by cr on 24.11.22.
//

#include "FWIAcousticInversionProb.hpp"
#include "Observation/FWIObservationOperatorDGTS.hpp"

double FWIVAcousticInversionProb::ut(double, const Cell &, const Point &, int) const { return 0; }

bool FWIVAcousticInversionProb::RHS(double) const { return true; }

void FWIVAcousticInversionProb::setInitialValueFlag(bool _flag) { initial_value = _flag; }

bool FWIVAcousticInversionProb::exact() const { return false; }

VectorField FWIVAcousticInversionProb::ForceV(double, const Cell &, const Point &, int) const { return zero; }

int FWIVAcousticInversionProb::getNrPointSrc() const { return Sources->GetNrSrces(); }

Point FWIVAcousticInversionProb::getPointSrcLoc(int nr) const { return Sources->GetShotLocation(nr); }

int FWIVAcousticInversionProb::GetShotCount() const { return Sources->GetShotCount(); }

void FWIVAcousticInversionProb::SetShot(int _shot) {
    if (_shot == -1) {
        Exit("dont call without shot number!")
    }
    if (Sources) Sources->SetShot(_shot);
}

double FWIVAcousticInversionProb::Kappa_i(const Cell &c, const Point &z, int i) const {
    return (i == 0) ? KappaRef(c, z) : Tau_p(c, z) * KappaRef(c, z);
}

double FWIVAcousticInversionProb::KappaRef(const Cell &C, const Point &z) const {
    double v = V_p(C, z);
    return (GetNumL() == 0) ? Rho(C, z) * v * v : Rho(C, z) * v * v / (1 + alpha(C, z) * Tau_p(C, z));
}

//double FWIVAcousticInversionProb::ForceP_i(double t, const Cell &c, const Point &z, int n, int) const {
//    return (n == 0) ? (*TimeSource)(t) * factor : 0.0;
//}

double FWIVAcousticInversionProb::alpha(const Cell &c, const Point &z) const {
    return alphaValue;
}

double FWIVAcousticInversionProb::Tau_i(const Cell &C, const Point &x, int i) const {
    return tau_i[i - 1];
}

void FWIVAcousticInversionProb::calculateAlphaValue() {
    for (int l = 0; l < GetNumL(); ++l) {
        alphaValue += Squared(wRef * tau_i[l]) / (1 + Squared(wRef * tau_i[l]));
    }
}
double FWIVAcousticInversionProb::alpha1() const {
  return alphaValue;
}
double FWIVAcousticInversionProb::alpha2()  const  {
  double alpha2 = 0;
  for (int l = 0; l < GetNumL(); ++l) {
    alpha2 += (wRef * tau_i[l]) / (1 + Squared(wRef * tau_i[l]));
  }
  return  alpha2;
}

void SourceInversionProblem(const std::vector<FWIDataContainerRange> &obs,
                            const std::vector<FWIDataContainerRange> &obsFwd,
                            std::shared_ptr<FWIVAcousticInversionProb> &prob,
                            const FWIObservationOperatorDGTS &obsOp, double eps,
                            double Kappa, int j) {
  for (int shot = 0; shot < prob->GetShotCount();++shot){
    prob->SetShot(shot);
    const Point& src = prob->getPointSrcLoc(0);
    SourceWaveletBase& Wavelet = prob->GetTimeSourceRef();
    SourceInversion(obs[shot].getSeismogram(), obsFwd[shot].getSeismogram(), src, Wavelet, &obsOp.GetSpaceOp(),
                    eps,Kappa);
    Wavelet.WriteToFile(Config::GetDataPath() + "InvSourcetime_iter_" +
                        std::to_string(j) + "shot_" + std::to_string(shot));
  }
}
