//
// Created by cr on 18.11.22.
//

#include "FWILinOpAdjRed.hpp"

void FWILinOpAdjRed::calculateAdjointIntern(Vector &AdjointVector, Vectors *LambdaFull) {
    assemble->ResetTime(TConf.t0, TConf.T, TConf.dt);
    std::unique_ptr<Vector> LambdaStep2;
    std::unique_ptr<Vector> LambdaStepUnpadded;
    if (PMLH) {
        LambdaStep2 = std::make_unique<ABCWaveVectorFWI>(*PMLH);
        LambdaStepUnpadded = std::make_unique<Vector>(U_m[0]);
        LambdaStepUnpadded->Clear();
    } else {
        LambdaStep2 = std::make_unique<Vector>(U_m[0]);
    }
    LambdaStep2->Clear();
    AdjointVector.Clear();
    timeIntegrator->Initialize(assemble.get(), *LambdaStep2);
    timeIntegrator->SetDivergenceCheck(false);
    mout.StartBlock("TI");
    mout << "Starting " << timeIntegrator->Name() << " (Order=" << timeIntegrator->GetOrder() << ")" << endl;
    while (!assemble->IsFinished()) {
        timeIntegrator->OneTimeStep(assemble.get(), *LambdaStep2);
        if (PMLH) {
            (*LambdaStepUnpadded).Clear();
            ABCWaveVectorFWI::transfer(PMLH->GetPMLToNonPML(), *LambdaStep2, *LambdaStepUnpadded);
            CalculateL2AdjointStep(*LambdaStepUnpadded, AdjointVector, assemble->Step());
            if (LambdaFull) (*LambdaFull)[assemble->Step()] = *LambdaStepUnpadded;
        } else {
          CalculateL2AdjointStep(*LambdaStep2, AdjointVector, assemble->Step());
          if (LambdaFull) (*LambdaFull)[assemble->Step()] = *LambdaStep2;
        }
    }
    timeIntegrator->Destruct();
    mout.EndBlock();
    AdjointVector *= -1.0;
}

FWILinOpAdjRed::FWILinOpAdjRed(const Vectors &U_m_, const Vector &MatVec_, const FWIObservationOperatorDGTS &Obs,
                               const std::shared_ptr<FWIVAcousticInversionProb> &prob_,
                               const std::shared_ptr<ABCHandler> &PMLH_, const TimeConf &TConf_,
                               const ReconstructionContainer &recon,
                               std::shared_ptr<EleCacheType> eleCache_) : PMLH(PMLH_),
                                                                                                      eleCache(
                                                                                                              std::move(eleCache_)),
                                                                                                      U_m(U_m_),
                                                                                                      MatVec(MatVec_),
                                                                                                      TConf(TConf_) {
    AdjointRHS = std::make_shared<DGVAcousticRHSAssemblerAdjoint>(Obs);
    assemble = std::make_shared<DGVAcousticAssembleCombined>(std::make_shared<DGVAcousticPlotter>(prob_->GetNumL(), nullptr),
                                                             std::make_shared<DGVAcousticMatrixAssembler>(std::make_unique<
                                            VAcousticMatFromVec>(*prob_, MatVec, PMLH.get()), false),
                                                             AdjointRHS, TConf.t0,
                                                             TConf.T,TConf.dt);
    if (PMLH) {
        assemble->SetEleCache(PMLH->GetEleCache().get());
    } else {
        assemble->SetEleCache(eleCache.get());
    }
    fwiAdjointCalculator = std::make_unique<FWIAdjointCalculator>(MatVec, prob_->getParametrization(), *eleCache,
                                                                  recon,
                                               prob_->GetNumL(), TConf.dt, prob_);
    timeIntegrator = GetFWITI(TConf.rkorder);
}

void FWILinOpAdjRed::CalculateL2AdjointStep(const Vector &LambdaStep, Vector &AdjointVector,
                                            int step) {
  int StepBwd = (U_m).size() - 1 - step;
    const Vector &U_m_bwd(U_m[StepBwd]);
    Vector U_m_fd(U_m[StepBwd + 1]);
    U_m_fd -= U_m_bwd;
    U_m_fd *= (1.0 / TConf.dt);
    fwiAdjointCalculator->calculateL2AdjointOneStep(U_m_bwd, U_m_fd, LambdaStep, AdjointVector);
}
