//
// Created by cr on 17/01/23.
//

#include <fftw3.h>
#include <complex>
#include "SourceInversion.hpp"
#include "Observation/FWIObservationOperatorSpaceDGTS.hpp"
std::pair<std::vector<double>,std::vector<double>>  ExtractRealAndIm(std::vector<std::complex<double>> &data) {
  std::vector<double> Real(data.size(),0.0);
  std::vector<double> Im(data.size(),0.0);
  for (int i =0; i<data.size();++i){
    Real[i] = data[i].real();
    Im[i]  = data[i].imag();
  }
  return {Real,Im};
}

void writeToASCII(std::string fn,std::vector<double> &data) {
  std::ofstream dataFile(fn);
  for(int i =1; i<data.size();++i){
    char s[100];
    sprintf(s, "%e\n",data[i]);
    dataFile << s;
  }
dataFile.close();
}
void SourceInversion(const DGSeismogram &obs, const DGSeismogram &fwd, const Point &srcPos, SourceWaveletBase &src,
                     const FWIObservationOperatorSpaceDGTS * SpOp, double eps, double kappa) {
  int ns = fwd.getSpecification().GetNumberOfTimeSteps();
  int Npad = (int) (pow(2.0,
                        ceil(logf((float) (fwd.getSpecification()
                                                  .GetNumberOfTimeSteps())) / log(2.0))));
  std::vector<std::complex<double>> sumn(Npad,0.0),sumd(Npad,0.0), D_s(Npad,0.0), D_ss(Npad,0.0);
  std::vector<std::complex<double>> D_s2(Npad,0.0), D_ss_fd(Npad,0.0), D_s_td(Npad,0.0);
  auto &wavelet =  src.GetEvalsOnGrid();
  for (int j = 0; j < ns-1; j++) {
      D_ss[j].real(wavelet[j+1]);
      D_ss[j].imag(0.0);
  }
  fftw_plan p_s;
  p_s = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(D_ss.data()),
                         reinterpret_cast<fftw_complex *>(D_ss_fd.data()), 1, FFTW_ESTIMATE);
  fftw_execute(p_s);
  double Ebar = 0.0;
  int ntr_glob = fwd.getSpecification().size();
  auto receiversSorted = fwd.getSpecification();
  std::sort(receiversSorted.begin(),receiversSorted.end());
  int i =0;
  for (const auto& rec : receiversSorted ) {
    std::vector<std::complex<double>> in_dat(Npad,0.0),out_dat(Npad,0.0),in_model(Npad,0.0),out_model(Npad,0.0);
    if (!SpOp->MeasureHere(rec)){
      ++i;
      continue;
    }
    double f_k = pow(dist(srcPos, rec), kappa);
    const auto &tsFwd = fwd.GetTimeSeries(rec);
    const auto &tsObs = obs.GetTimeSeries(rec);
    fftw_plan p_dat, p_model;
    for (int j = 1; j < Npad; j++) {
      if (j < tsFwd.size()-1) {
        in_model[j].real(tsFwd[j+1]);
        in_model[j].imag(0.0);
        in_dat[j].real(tsObs[j+1]);
        in_dat[j].imag(0);
      } else {
        in_model[j].real(0);
        in_model[j].imag(0);
        in_dat[j].real(0);
        in_dat[j].imag(0);
      }
    }
    p_dat = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(in_dat.data()),
                             reinterpret_cast<fftw_complex *>(out_dat.data()), 1, FFTW_ESTIMATE);
    p_model = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(in_model.data()),
                               reinterpret_cast<fftw_complex *>(out_model.data()), 1, FFTW_ESTIMATE);
    fftw_execute(p_dat);
    fftw_execute(p_model);
    for (int j = 0; j < Npad; j++) {
      sumn[j] +=  ntr_glob * out_dat[j] * conj(out_model[j]);
      sumd[j] +=  ntr_glob * out_model[j] * conj(out_model[j]);
      Ebar += (out_model[j] * conj(out_model[j])).real();
    }
    fftw_destroy_plan(p_dat);
    fftw_destroy_plan(p_model);
    ++i;
  }
  PPM->Sum(sumn);
  PPM->Sum(sumd);
  Ebar= PPM->Sum(Ebar);
  Ebar = Ebar / Npad;
  for (int j = 0; j < Npad; j++) {
    D_s[j] = sumn[j]/(sumd[j].real() + eps * eps* Ebar);
  }
  fftw_plan p_cf;
  p_cf = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(D_s.data()),
                          reinterpret_cast<fftw_complex *>(D_s_td.data()), -1, FFTW_ESTIMATE);
  fftw_execute(p_cf);
  for (int j = ns; j < Npad; j++) {
    D_s_td[j].real(0.0);
    D_s_td[j].imag(0.0);
  }
  fftw_plan p_cf2;
  p_cf2 = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(D_s_td.data()),
                          reinterpret_cast<fftw_complex *>(D_s.data()), 1, FFTW_ESTIMATE);
  fftw_execute(p_cf2);
  for (int j = 0; j < Npad; j++) {
    D_s[j] = D_ss_fd[j] * D_s[j]* 1.0/Npad;
  }
  fftw_destroy_plan(p_cf);
  fftw_destroy_plan(p_cf2);
  fftw_plan p_stf;
  p_stf = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(D_s.data()),
                           reinterpret_cast<fftw_complex *>(D_s_td.data()), -1, FFTW_ESTIMATE);
  fftw_execute(p_stf);
  RVector waveletTmp(wavelet);
  waveletTmp = 0.0;
  for (int j = 0; j < wavelet.size()-1;++j){
    waveletTmp[j+1] = D_s_td[j].real()*1.0/Npad;;
  }
  fftw_destroy_plan(p_stf);
  fftw_destroy_plan(p_s);
  src.SetEvalsOnGrid(waveletTmp);
}