//
// Created by cr on 03/03/23.
//

#include "FWITestSTI.hpp"

#include <memory>

#include <gtest/gtest.h>

#include "DGVAcousticAssembleCombined.hpp"
#include "ForwardModels/ForwardModel.hpp"
#include "MeshIndex.hpp"
#include "Observation/DGSeismogram.hpp"
#include "Observation/FWIObservationOperatorDGTS.hpp"
#include "Observation/ObservationCreator.hpp"
#include "Problems/FWIInversionProblems.hpp"
#include "Problems/Sources/FWISourceWavelet.hpp"
#include "SourceInversion.hpp"
#include "TestEnvironment.hpp"
struct TestParams {

  std::string srcSignal;
  double eps{};
};
class FWTTestSTI : public TestWithParam<TestParams> {
protected:
  std::unique_ptr<DGSeismogram> FwdSeis;
  std::unique_ptr<SourceWaveletBase> FwdWavelet;
public:
  FWTTestSTI() {
    Config::Initialize(FWITestSTI);
    auto meshes = getFWIMesh();

    WPConf Conf;
    MeshIndex lev{Conf.level, -1};
    int deg = 0;
    Config::Get("deg",deg);
    std::string SignalName = "Sin";
    Conf.SConf.SignalName =SignalName;
    FwdWavelet = GetSourceWavelet(Conf.TConf, Conf.SConf);
    Conf.s_prob = "VAcousticWaveC2Benchmark";
    std::shared_ptr<FWIVAcousticInversionProb> problem = GetInversionProblem(Conf);
    FWIDiscretizationStack Stack(meshes, problem->GetParamComps(),
                                 problem->GetStateComps(), deg,
                                 lev);
    (*Stack.wD)().PrintInfo();
    auto eleCache = std::make_unique<EleCacheContainer>(*Stack.wV, problem->GetNumL());
    auto spec = ObservationSpecificationBuilder().WithT(Conf.TConf.T).
                Withdt(Conf.TConf.dt).Witht0(Conf.TConf.t0).CreateShared();
    FWIObservationOperatorDGTS *Obs =
        ObservationCreator(eleCache->Get(),
                           Stack.GetMesh(),
                           problem->GetNumL()).WithSpec(
                spec).Create();
    std::cout << std::setprecision(15);
    problem->setInitialValueFlag(true);
    auto M = std::make_unique<Vector>(*Stack.mV);
    auto M2 = std::make_unique<Vector>(*Stack.mV);
    M2->Clear();
    M->Clear();
    FillVectorFromProblem(*problem, *M2);
    ReparametrizeMaterialVector(*M2, *problem, *M);
    std::shared_ptr<ABCHandler> PMLH = nullptr;
    Stack.wV->PrintInfo();
    Stack.GetMesh().PrintInfo();
//    Exit("CIAO")
    problem->SetShot(0);
    FWITimeSolver TSolveR(*Stack.wV, PMLH, Conf.TConf.rkorder);
    auto Mat = std::make_unique<VAcousticMatFromVec>(*problem, *M2);
    auto Mat2 = std::make_unique<VAcousticMatFromVec>(*problem, *M2);
    auto MatAss = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), true, eleCache->Get().get());
    auto plotter = std::make_shared<DGVAcousticPlotter>(problem->GetNumL(), eleCache->Get().get());
    std::shared_ptr<FWIProblemWrapper> problemVA = std::make_shared<FWIProblemWrapper>(*problem, *problem,
                                                                                       std::move(M),
                                                                                       problem->GetNumL(),
                                                                                       problem->GetDim(),
                                                                                       problem);

    auto rhsAssembler = std::make_shared<DGVAcousticRHSAssemblerDefault>(*problemVA);
    auto assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(plotter, MatAss, rhsAssembler);
    assembleCombined->SetEleCache(eleCache->Get().get());
    FwdSeis = std::make_unique<DGSeismogram>(Obs->GetSpaceOp().GetParaDat());
//    FwdSeis->ReadFromFile("CONVERT_SIN");
    TSolveR.SolveAndMeasure(assembleCombined,*Obs, *FwdSeis, nullptr);
    problem->GetTimeSourceRef().WriteToFile("SourceWavelet_FWD_"+ SignalName);
    FwdSeis->WriteToFile(SignalName +"Seismogram_FWD");
  }

  void TearDown() override {
    Config::Close();
  }
};


TEST_P(FWTTestSTI, FWTTestSTI) {
  auto meshes = getFWIMesh();
  WPConf Conf;
  MeshIndex lev{Conf.level, -1};
  int deg = 0;
  int numL = 0;
  Conf.SConf.SignalName =GetParam().srcSignal;
  Config::Get("deg",deg);

  Conf.s_prob = "VAcousticWaveC2Benchmark";
  std::shared_ptr<FWIVAcousticInversionProb> problem = GetInversionProblem(Conf);
  FWIDiscretizationStack Stack(meshes, problem->GetParamComps(),
                               problem->GetStateComps(), deg,
                               lev);
  (*Stack.wD)().PrintInfo();
  auto eleCache = std::make_unique<EleCacheContainer>(*Stack.wV, problem->GetNumL());
  auto spec = ObservationSpecificationBuilder().WithT(Conf.TConf.T).
              Withdt(Conf.TConf.dt).Witht0(Conf.TConf.t0).CreateShared();
  FWIObservationOperatorDGTS *Obs =
      ObservationCreator(eleCache->Get(),
                         Stack.GetMesh(),
                         problem->GetNumL()).WithSpec(
              spec).Create();
  std::cout << std::setprecision(15);
  problem->setInitialValueFlag(true);
  auto M = std::make_unique<Vector>(*Stack.mV);
  auto M2 = std::make_unique<Vector>(*Stack.mV);
  M2->Clear();
  M->Clear();
  FillVectorFromProblem(*problem, *M2);
  ReparametrizeMaterialVector(*M2, *problem, *M);
  std::shared_ptr<ABCHandler> PMLH = nullptr;

  problem->SetShot(0);
  FWITimeSolver TSolveR(*Stack.wV, PMLH, Conf.TConf.rkorder);
  auto Mat = std::make_unique<VAcousticMatFromVec>(*problem, *M2);
  auto Mat2 = std::make_unique<VAcousticMatFromVec>(*problem, *M2);
  auto MatAss = std::make_shared<DGVAcousticMatrixAssembler>(std::move(Mat), true, eleCache->Get().get());
  auto plotter = std::make_shared<DGVAcousticPlotter>(problem->GetNumL(), eleCache->Get().get());
  std::shared_ptr<FWIProblemWrapper> problemVA = std::make_shared<FWIProblemWrapper>(*problem, *problem,
                                                                                     std::move(M),
                                                                                     problem->GetNumL(),
                                                                                     problem->GetDim(),
                                                                                     problem);

  auto rhsAssembler = std::make_shared<DGVAcousticRHSAssemblerDefault>(*problemVA);
  auto assembleCombined = std::make_shared<DGVAcousticAssembleCombined>(plotter, MatAss, rhsAssembler);
  assembleCombined->SetEleCache(eleCache->Get().get());
  DGSeismogram ObsSeis(Obs->GetSpaceOp().GetParaDat());
  TSolveR.SolveAndMeasure(assembleCombined,*Obs, ObsSeis, nullptr);
//  ObsSeis.ReadFromFile("CONVERT");
  const auto &EvalsUsedSource =  problem->GetTimeSourceRef().GetEvalsOnGrid();
  problem->GetTimeSourceRef().WriteToFile("SourceWavelet_"+GetParam().srcSignal);
  Point dummy(750.0,1000.0);
  const auto &EvalsRealSource = FwdWavelet->GetEvalsOnGrid();
  SourceInversion(ObsSeis, *FwdSeis, dummy,*FwdWavelet,
                       &Obs->GetSpaceOp(), GetParam().eps,1);
  double err;
  double energy;
  for (int i =0 ; i< EvalsRealSource.size(); ++i){
    err += (EvalsRealSource[i]- EvalsUsedSource[i])*(EvalsRealSource[i]- EvalsUsedSource[i]);
    energy += EvalsRealSource[i]*EvalsRealSource[i];
  }
  MOUT(err);
  MOUT(energy);
  MOUT(err/energy);
  ObsSeis.WriteToFile(GetParam().srcSignal + "Seismogram");
  FwdWavelet->WriteToFile("ReconstructedWavelet_" + GetParam().srcSignal);
}
INSTANTIATE_TEST_SUITE_P(FWTTestSTI, FWTTestSTI, Values(TestParams{"Ricker",0.005},
                         TestParams{"RickerDerivative",0.005}));


int main(int argc, char **argv) {
  return MppTest(
      MppTestBuilder(argc, argv).
          WithoutDefaultConfig().
          WithScreenLogging().
          WithConfPath("../").
          WithGeoPath("../conf/geo/").
          WithPlotPath("data/vtu/").
          WithPPM()
  ).RUN_ALL_MPP_TESTS();
}