//
// Created by cr on 22.08.22.
//

#include "DGVAcousticRHSAssembler.hpp"

void DGVAcousticRHSAssemblerDefault::assemblePointSrcRHSPart(double t, Vector &b) const {
    for (int nr = 0; nr < problem.getNrPointSrc(); ++nr) {
        const Point src_loc = problem.getPointSrcLoc(nr);
        int size = pointSrcCells.find(src_loc) == pointSrcCells.end() ? 0 : pointSrcCells.at(src_loc).size();
        const unsigned int cell_count = PPM->Sum(size);
        if (pointSrcCells.find(src_loc) == pointSrcCells.end()) {
            continue;
        }
        for (std::unique_ptr<cell> &cell: pointSrcCells.at(src_loc)) {
            const Cell &C = *(*cell);
            DGAcousticElement *elem = EleCache == nullptr ? new DGAcousticElement(b, C, problem.NumDamping())
                                                          : EleCache->find(C())->second;
            int p_dim = elem->p_dimension();
            Point cmid = C();
            int nV = elem->V_dimension();
            int nP = elem->shape_size();

            Point lz = C.GlobalToLocal(src_loc);
            VectorField V_rhs(0.0);
            RVector ForceP(0.0, p_dim);
            problem.Force(V_rhs, ForceP, t, *(*cell), src_loc);
            ForceP *= 1.0 / cell_count;
            V_rhs *= 1.0 / cell_count;
            for (int i = 0; i < elem->shape_size(); ++i) {
                double P_i = elem->evaluateShape(lz, i);
                for (int n = 0; n < p_dim; ++n) {
                    double factor = matInRHS ? 1.0 / problem.Kappa_i(*(*cell), (*cell)(), 0) : 1;
//                    double factor = 1;
                    b((*cell)(), nV + n * nP + i) += P_i * ForceP[n] * factor;
                }
            }
            for (int i = 0; i < nV; ++i) {
                VectorField V_i = elem->evaluateVelocityShape(lz, i);
                b((*cell)(), i) += (V_i * V_rhs);
            }
        }
    }
}

DGAcousticElement *DGVAcousticRHSAssemblerDefault::getElem(const VectorMatrixBase &base, const Cell &C) const {
    if (EleCache == nullptr)
        return new DGAcousticElement(base, C, problem.NumDamping());
    else {
        auto p = EleCache->find(C());
        return p->second;
    }
}

void DGVAcousticRHSAssemblerDefault::destroyElement(DGAcousticElement *elem) const {
    if (EleCache == nullptr) {
        delete elem;
    }
}

void
DGVAcousticRHSAssemblerDefault::assembleRHSPart(double t, Vector &b, const cell &c,
                                                const DGAcousticElement *elem) const {
    int p_dim = elem->p_dimension();
    int Row = b.find_row(c()).Id();
    for (int q = 0; q < elem->nQ(); ++q) {
        double w = elem->QWeight(q);
        int nV = elem->V_dimension();
        int nP = elem->shape_size();
        Point z = elem->QPoint(q);
        VectorField V_rhs(0.0);
        RVector ForceP(0.0, p_dim);
        problem.Force(V_rhs, ForceP, t, *c, z);
        for (int i = 0; i < elem->V_dimension(); ++i) {
            VectorField V_i = elem->Velocity(q, i);
            b(Row, i) += w * V_i * V_rhs;
        }
        for (int i = 0; i < elem->shape_size(); ++i) {
            double P_i = elem->Pressure(q, i);
            for (int n = 0; n < p_dim; ++n) {
//                        double factor = MatInRHS ? 1.0 / problem.Kappa_i(c, c(), 0) : 1;
                double factor = 1.0;
                b(Row, nV + n * nP + i) += w * factor * P_i * ForceP[n];
            }
        }
    }
}

void
DGVAcousticRHSAssemblerDefault::assembleExactRHSPart(double t, Vector &b, const cell &c, const DGAcousticElement *elem,
                                                     bool fwd) const {
    Exit("debug this again");
    int p_dim = elem->p_dimension();
    int Row = b.find_row(c()).Id();
    for (int f = 0; f < c.Faces(); ++f) {
        if (!b.OnBoundary(*c, f)) continue;
        const Cell &C = *c;
        DGAcousticFaceElementT
                felem(b, C, f, problem.NumDamping());
        int nV = felem.V_dimension();
        int nP = felem.shape_size();
        BFParts bnd(b.GetMesh(), c);
        for (int q = 0; q < felem.nQ(); ++q) {
            double w = felem.QWeight(q);
            VectorField N = felem.QNormal(q);
            Point z = felem.QPoint(q);
            double Z_L = sqrt(problem.Kappa(*c, z)
                              * problem.Rho(*c, z));
            double alpha = 1 / Z_L;
            if (bnd[f] == 2) {
              double VN_bnd = problem.Neumann(t, *c, z, N);
              for (int i = 0; i < nV; ++i) {
                double VN_i = felem.Velocity(q, i) * N;
                b(Row, i) += w * Z_L * VN_i * VN_bnd;
              }
                for (int i = 0; i < nP; ++i) {
                    double P_i = felem.Pressure(q, i);
                    for (int n = 0; n < p_dim; ++n) {
                        b(Row, nV + n * nP + i) += w * P_i * VN_bnd;
                    }
                }
            } else if (bnd[f] == 1) {
                for (int i = 0; i < nV; ++i) {
                  double VN_i = felem.Velocity(q, i) * N;
                  for (int n = 0; n < p_dim; ++n) {
                    b(Row, i) += w * VN_i * problem.Dirichlet_i(t, *c, z, n);
                    }
                }
                for (int i = 0; i < nP; ++i) {
                    double P_i = felem.Pressure(q, i);
                    for (int n = 0; n < p_dim; ++n) {
                        for (int k = 0; k < p_dim; ++k) {
                            b(c(), nV + n * nP + i) += w * alpha * P_i
                                                       * problem.Dirichlet_i(t, *c, z, k);
                        }
                    }
                }
            } else {
                Exit(("boundary type " + std::to_string(bnd[f]))
                             .c_str());
            }
        }
    }
}

void DGVAcousticRHSAssemblerDefault::initializePointSrcCache(Vector &b) const {
    for (cell c = b.cells(); c != b.cells_end(); ++c) {
        for (int nr = 0; nr < problem.getNrPointSrc(); ++nr) {
            const Point &src_loc = problem.getPointSrcLoc(nr);
          if (c.PointInCell(src_loc)) {
            pointSrcCells[src_loc].push_back(std::unique_ptr<cell>(new cell(c)));
            }
        }
    }
}

void DGVAcousticRHSAssembler::SetElementCache(const std::unordered_map<Point, DGAcousticElement *> *_EleCache) {
    EleCache = _EleCache;
}

void DGVAcousticRHSAssemblerDefault::RHS(double t, Vector &b) const {
    b.Clear();
    if (!problem.RHS(t)) return;
    Initialize(b);
    if (problem.getNrPointSrc() > 0 && problem.getNrPointSrc() < INT_MAX) {
        assemblePointSrcRHSPart(t, b);
    } else {
      for (cell c = b.cells(); c != b.cells_end(); ++c) {
            const Cell &C = *c;
            DGAcousticElement *elem = getElem(b, C);
            assembleRHSPart(t, b, c, elem);
            if (problem.HasExactSolution()) {
                assembleExactRHSPart(t, b, c, elem, true);
            }
            destroyElement(elem);
        }
    }
}

void DGVAcousticRHSAssemblerDefault::Initialize(Vector &u) const {
    int size_list = pointSrcCells.begin() == pointSrcCells.end() ? 0 : pointSrcCells.size();
    const unsigned int size_list_sum = PPM->Sum(size_list);
    if (size_list_sum == 0 && problem.getNrPointSrc() > 0) { initializePointSrcCache(u);
    }
}
